"""Utilities for parameter validation."""

# Authors: Zhining Liu <zhining.liu@outlook.com>
# License: MIT


from copy import copy
from warnings import warn
from collections import Counter
from inspect import signature

import numbers
import numpy as np
from math import ceil
from sklearn.ensemble import BaseEnsemble
from sklearn.utils import check_array, check_X_y
from sklearn.utils.validation import check_is_fitted
from sklearn.metrics import accuracy_score, balanced_accuracy_score, f1_score


SAMPLING_KIND = (
    "over-sampling",
    "under-sampling",
    "multi-class-hybrid-sampling",
)
SamplingKindError = NotImplementedError(
    f"'sampling_type' must be one of {SAMPLING_KIND}."
)


def _target_samples_int(y, n_target_samples, sampling_type):
    target_stats = dict(Counter(y))
    max_class_ = max(target_stats, key=target_stats.get)
    min_class_ = min(target_stats, key=target_stats.get)
    n_max_class_samples_ = target_stats[max_class_]
    n_min_class_samples_ = target_stats[min_class_]
    if sampling_type == 'under-sampling':
        if n_target_samples >= n_max_class_samples_:
            raise ValueError(
                f"'n_target_samples' >= the number of samples"
                f" of the largest class ({n_max_class_samples_})."
                f" Set 'n_target_samples' < {n_max_class_samples_}"
                f" to perform under-sampling properly."
            )
        target_distr = dict([
            (label, min(n_target_samples, target_stats[label]))
            for label in target_stats.keys()
        ])
        return target_distr
    elif sampling_type == 'over-sampling':
        if n_target_samples <= n_min_class_samples_:
            raise ValueError(
                f"'n_target_samples' <= the number of samples"
                f" of the largest class ({n_min_class_samples_})."
                f" Set 'n_target_samples' > {n_min_class_samples_}"
                f" to perform over-sampling properly."
            )
        target_distr = dict([
            (label, max(n_target_samples, target_stats[label]))
            for label in target_stats.keys()
        ])
        return target_distr
    elif sampling_type == "multi-class-hybrid-sampling":
        warning_info =  f" Set 'n_target_samples' between [{n_min_class_samples_}" + \
                f" , {n_max_class_samples_}] if you want to perform" + \
                f" multi-class hybrid-sampling (under-sample the minority" + \
                f" classes, over-sample the majority classes) properly."
        if n_target_samples >= n_max_class_samples_:
            raise Warning(
                f"'n_target_samples' >= the number of samples" + \
                f" of the largest class ({n_max_class_samples_})." + \
                f" ONLY over-sampling will be applied to all classes." + warning_info
            )
        elif n_target_samples <= n_min_class_samples_:
            raise Warning(
                f"'n_target_samples' <= the number of samples" + \
                f" of the largest class ({n_min_class_samples_})." + \
                f" ONLY under-sampling will be applied to all classes." + warning_info
            )
        target_distr = dict([
            (label, n_target_samples)
            for label in target_stats.keys()
        ])
        return target_distr
    else: raise SamplingKindError


def _target_samples_dict(y, n_target_samples, sampling_type):
    target_stats = dict(Counter(y))
    # check that all keys in n_target_samples are also in y
    set_diff_sampling_strategy_target = set(n_target_samples.keys()) - set(
        target_stats.keys()
    )
    if len(set_diff_sampling_strategy_target) > 0:
        raise ValueError(
            f"The {set_diff_sampling_strategy_target} target class is/are not "
            f"present in the data."
        )
    # check that there is no negative number
    if any(n_samples < 0 for n_samples in n_target_samples.values()):
        raise ValueError(
            f"The number of samples in a class cannot be negative."
            f"'n_target_samples' contains some negative value: {n_target_samples}"
        )

    if sampling_type == 'under-sampling':
        target_distr = copy(target_stats)
        for class_label, n_target_sample in n_target_samples.items():
            n_origin_sample = target_stats[class_label]
            if n_target_sample > n_origin_sample:
                raise ValueError(
                    f" The target number of samples of class {class_label}"
                    f" should be < {n_origin_sample} (number of samples"
                    f" in class {class_label}) to perform under-sampling,"
                    f" got {n_target_sample}."
                )
            else:
                target_distr[class_label] = n_target_sample
        return target_distr

    elif sampling_type == 'over-sampling':
        target_distr = copy(target_stats)
        for class_label, n_target_sample in n_target_samples.items():
            n_origin_sample = target_stats[class_label]
            if n_target_sample < n_origin_sample:
                raise ValueError(
                    f" The target number of samples of class {class_label}"
                    f" should be > {n_origin_sample} (number of samples"
                    f" in class {class_label}) to perform over-sampling,"
                    f" got {n_target_sample}."
                )
            else:
                target_distr[class_label] = n_target_sample
        return target_distr

    elif sampling_type == "multi-class-hybrid-sampling":
        target_distr = copy(target_stats)
        if all(n_target_samples[label] <= target_stats[label] for label in n_target_samples.keys()):
            raise Warning(
                f"The target number of samples is smaller than the number"
                f" of original samples for all classes. ONLY under-sampling"
                f" will be carried out."
            )
        elif all(n_target_samples[label] >= target_stats[label] for label in n_target_samples.keys()):
            raise Warning(
                f"The target number of samples is greater than the number"
                f" of original samples for all classes. ONLY over-sampling"
                f" will be carried out."
            )
        target_distr.update(n_target_samples)
        return target_distr
    
    else: raise SamplingKindError


def check_n_target_samples(y, n_target_samples, sampling_type):
    if isinstance(n_target_samples, numbers.Integral):
        return _target_samples_int(y, n_target_samples, sampling_type)
    elif isinstance(n_target_samples, dict):
        return _target_samples_dict(y, n_target_samples, sampling_type)
    else: raise ValueError(
        f"'n_target_samples' should be of type `int` or `dict`,"
        f" got {type(n_target_samples)}."
    )


def check_target_label(y, target_label, sampling_type):
    """check parameter `target_label`."""
    
    target_stats = dict(Counter(y))
    if isinstance(target_label, numbers.Integral):
        if target_label in target_stats.keys():
            return target_label
        else: raise ValueError(
            f"The target class {target_label} is not present in the data."
        )
    else: raise TypeError(
        f"'target_label' should be of type `int`,"
        f" got {type(target_label)}."
    )


def check_target_label_and_n_target_samples(y, target_label, n_target_samples, sampling_type):
    """Jointly check `target_label` and `n_target_samples` parameters."""

    # Store the original target class distribution
    target_stats = dict(Counter(y))
    min_class = min(target_stats, key=target_stats.get)
    maj_class = max(target_stats, key=target_stats.get)
    
    # if n_target_samples is NOT specified
    if n_target_samples == None:
        # Set target_label if NOT specified
        if target_label == None:
            if sampling_type == "under-sampling":
                target_label_ = min_class
            elif sampling_type == "over-sampling":
                target_label_ = maj_class
            elif sampling_type == "multi-class-hybrid-sampling":
                raise ValueError(
                    f"For \"multi-class-hybrid-sampling\", must specify"
                    f" 'n_target_samples' or 'target_label'."
                )    
            else: raise SamplingKindError
        # Check target_label
        else: target_label_ = check_target_label(y, target_label, sampling_type)
        # Set n_target_samples
        n_target_samples = target_stats[target_label_]
    
    # if n_target_samples is specified
    else:
        if target_label == None:
            target_label_ = target_label
        # n_target_samples and target_label CANNOT both be both specified
        else:
            raise ValueError(
                f"'n_target_samples' and 'target_label' cannot"
                f" be specified at the same time."
            )
    
    # Check n_target_samples
    target_distr_ = check_n_target_samples(y, n_target_samples, sampling_type)

    return target_label_, target_distr_


BALANCING_SCHEDULE_PARAMS_TYPE = {
    'origin_distr': dict,
    'target_distr': dict,
    'i_estimator': numbers.Integral,
    'total_estimator': numbers.Integral,
}


def _uniform_schedule(origin_distr, target_distr, i_estimator, total_estimator):
    """Return target distribution"""
    for param, (param_name, param_type) in zip(
            [origin_distr, target_distr, i_estimator, total_estimator], 
            list(BALANCING_SCHEDULE_PARAMS_TYPE.items())):
        if not isinstance(param, param_type):
            raise TypeError(
                f"'{param_name}' must be `{param_type}`, got {type(param)}."
            )
    if i_estimator >= total_estimator:
        raise ValueError(
            f"'i_estimator' should < 'total_estimator',"
            f" got 'i_estimator' = {i_estimator} >= 'total_estimator' = {total_estimator}."
        )
    return target_distr


def _progressive_schedule(origin_distr, target_distr, i_estimator, total_estimator):
    """Progressively interpolate between original and target distribution"""
    for param, (param_name, param_type) in zip(
            [origin_distr, target_distr, i_estimator, total_estimator], 
            list(BALANCING_SCHEDULE_PARAMS_TYPE.items())):
        if not isinstance(param, param_type):
            raise TypeError(
                f"'{param_name}' must be `{param_type}`, got {type(param)}."
            )
    if i_estimator >= total_estimator:
        raise ValueError(
            f"'i_estimator' should < 'total_estimator',"
            f" got 'i_estimator' = {i_estimator} >= 'total_estimator' = {total_estimator}."
        )
    result_distr = {}
    if total_estimator == 1:
        progress_ = 1
    else: progress_ = i_estimator / (total_estimator-1)
    for label in origin_distr.keys():
        result_distr[label] = ceil(
            origin_distr[label]*(1.-progress_) + \
            target_distr[label]*progress_ - 1e-10
        )
    return result_distr


BALANCING_KIND_MAPPING = {
    "uniform": _uniform_schedule,
    "progressive": _progressive_schedule,
}
BALANCING_KIND = list(BALANCING_KIND_MAPPING.keys())
BALANCING_SCHEDULE_INFO = \
    "\nNote: self-defined `balancing_schedule` should take 4 positional" + \
    " arguments with order ('origin_distr': `dict`, 'target_distr':" + \
    " `dict`, 'i_estimator': `int`, 'total_estimator': `int`), and" + \
    " return a 'result_distr': `dict`. For all `dict`, the keys" + \
    " correspond to the targeted classes, and the values correspond to the" + \
    " (desired) number of samples for each class."


def check_balancing_schedule(balancing_schedule):
    """Check the `balancing_schedule` parameter."""
    if callable(balancing_schedule):
        try:
            return_value = balancing_schedule({}, {}, 0, 0)
        except Exception as e:
            e_args = list(e.args)
            e_args[0] += BALANCING_SCHEDULE_INFO
            e.args = tuple(e_args)
            raise e
        else:
            if not isinstance(return_value, dict):
                raise TypeError(
                    f" The self-defined `balancing_schedule` must return a `dict`," + \
                    f" got {type(return_value)}" + \
                    BALANCING_SCHEDULE_INFO
                )
        return balancing_schedule

    if balancing_schedule in BALANCING_KIND:
        return BALANCING_KIND_MAPPING[balancing_schedule]
    else:
        balancing_schedule_info = balancing_schedule if isinstance(balancing_schedule, str) \
            else type(balancing_schedule)
        raise TypeError(
            f"'balancing_schedule' should be one of {BALANCING_KIND} or `callable`,"
            f" got {balancing_schedule_info}."
        )


EVAL_METRICS_DEFAULT = {
    'acc': (accuracy_score, {}),
    'balanced_acc': (balanced_accuracy_score, {}),
    'weighted_f1': (f1_score, {'average':'weighted'}),
}
EVAL_METRICS_INFO = \
            " Example 'eval_metrics': {..., 'metric_name': ('metric_func', 'metric_kwargs'), ...}."
            # " where `metric_name` is string, `metric_func` is `callable`," + \
            # " and `metric_arguments` is a dict of arguments" + \
            # " that needs to be passed to the metric function," + \
            # " e.g., {..., `argument_name`: `value`}."
EVAL_METRICS_TUPLE_TYPE = {
    'metric_func': callable,
    'metric_kwargs': dict,
}
EVAL_METRICS_TUPLE_LEN = len(EVAL_METRICS_TUPLE_TYPE)


def _check_eval_metric_func(metric_func):
    if not callable(metric_func):
        raise TypeError(
            f" The 'metric_func' should be `callable`, got {type(metric_func)},"
            f" please check your usage."
            + EVAL_METRICS_INFO
        )
    if 'y_true' not in signature(metric_func).parameters:
        raise RuntimeError(
            f"The metric function must have the keyword argument 'y_true'"
            f" (true labels or binary label indicators, 1d-array of shape (n_samples,))."
        )
    if 'y_pred' not in signature(metric_func).parameters and \
        'y_score' not in signature(metric_func).parameters:
        raise RuntimeError(
            f"The metric function must have the keyword argument 'y_pred' or 'y_score'."
            f" When use 'y_pred': it corresponds to predicted labels, 1d-array of shape (n_samples,)."
            f" When use 'y_score': it corresponds to predicted labels, or an array of shape"
            f" (n_samples, n_classes) of probability estimates provided by the predict_proba method.)"
        )
    accept_proba = 'y_score' in signature(metric_func).parameters
    accept_labels = 'labels' in signature(metric_func).parameters
    return metric_func, accept_proba, accept_labels


def _check_eval_metric_args(metric_kwargs):
    if not isinstance(metric_kwargs, dict):
        raise TypeError(
            f" The 'metric_kwargs' should be a `dict` of arguments"
            f" that needs to be passed to the metric function,"
            f" got {type(metric_kwargs)}, "
            f" please check your usage."
            + EVAL_METRICS_INFO
        )
    return metric_kwargs


def _check_eval_metric_name(metric_name):
    if not isinstance(metric_name, str):
        raise TypeError(
            f" The keys must be `string`, got {type(metric_name)}, "
            f" please check your usage."
            + EVAL_METRICS_INFO
        )
    return metric_name


def _check_eval_metric_tuple(metric_tuple, metric_name):
    if not isinstance(metric_tuple, tuple):
        raise TypeError(
            f" The value of '{metric_name}' is {type(metric_tuple)} (should be tuple)," + \
            f" please check your usage."
            + EVAL_METRICS_INFO
        )
    elif len(metric_tuple) != EVAL_METRICS_TUPLE_LEN:
        raise ValueError(
            f" The data tuple of '{metric_name}' has {len(metric_tuple)} element(s)" + \
            f" (should be {EVAL_METRICS_TUPLE_LEN}), please check your usage."
            + EVAL_METRICS_INFO
        )
    else:
        metric_func_, accept_proba, accept_labels = _check_eval_metric_func(metric_tuple[0])
        metric_kwargs_ = _check_eval_metric_args(metric_tuple[1])
        return (
            metric_func_,
            metric_kwargs_,
            accept_proba,
            accept_labels,
        )


def _check_eval_metrics_dict(eval_metrics_dict):
    """check 'eval_metrics' dict."""
    eval_metrics_dict_ = {}
    for metric_name, metric_tuple in eval_metrics_dict.items():
        
        metric_name_ = _check_eval_metric_name(metric_name)
        metric_tuple_ = _check_eval_metric_tuple(metric_tuple, metric_name_)
        eval_metrics_dict_[metric_name_] = metric_tuple_
    
    return eval_metrics_dict_

def check_eval_metrics(eval_metrics):
    """Check parameter `eval_metrics`."""
    if eval_metrics is None:
        return _check_eval_metrics_dict(EVAL_METRICS_DEFAULT)
    elif isinstance(eval_metrics, dict):
        return _check_eval_metrics_dict(eval_metrics)
    else: 
        raise TypeError(
            f"'eval_metrics' must be of type `dict`, got {type(eval_metrics)}, please check your usage."
            + EVAL_METRICS_INFO
        )


TRAIN_VERBOSE_TYPE = {
    'granularity': numbers.Integral,
    'print_distribution': bool,
    'print_metrics': bool,
}

TRAIN_VERBOSE_DEFAULT = {
    # 'granularity' will be set to int(n_estimators_ensemble/10)
    #  when check_train_verbose() is called
    'print_distribution': True,
    'print_metrics': True,
}

TRAIN_VERBOSE_DICT_INFO = \
        " When 'train_verbose' is `dict`, at least one of the following" + \
        " terms should be specified: " + \
        " {'granularity': `int` (default=1)," + \
        " 'print_distribution': `bool` (default=True)," + \
        " 'print_metrics': `bool` (default=True)}."


def check_train_verbose(train_verbose:bool or numbers.Integral or dict,
                        n_estimators_ensemble:int, training_type:str, 
                        **ignored_properties):
                        # n_estimators_ensemble:int,):

    train_verbose_ = copy(TRAIN_VERBOSE_DEFAULT)
    train_verbose_.update({
        'granularity': max(1, int(n_estimators_ensemble/10))
    })

    if training_type == 'parallel':
        # For ensemble classifiers trained in parallel
        # train_verbose can only be of type bool
        if isinstance(train_verbose, bool):
            if train_verbose == True:
                train_verbose_['print_distribution'] = False 
                return train_verbose_
            if train_verbose == False:
                return False
        else: raise TypeError(
            f"'train_verbose' can only be of type `bool`"
            f" for ensemble classifiers trained in parallel,"
            f" gor {type(train_verbose)}."
        )
    
    elif training_type == 'iterative':
        # For ensemble classifiers trained in iterative manner
        # train_verbose can be of type bool / int / dict
        if isinstance(train_verbose, bool):
            if train_verbose == True:
                return train_verbose_
            if train_verbose == False:
                return False

        if isinstance(train_verbose, numbers.Integral):
            train_verbose_.update({'granularity': train_verbose})
            return train_verbose_
            
        if isinstance(train_verbose, dict):
            # check key value type
            set_diff_verbose_keys = set(train_verbose.keys()) - set(TRAIN_VERBOSE_TYPE.keys())
            if len(set_diff_verbose_keys) > 0:
                raise ValueError(
                    f"'train_verbose' keys {set_diff_verbose_keys} are not supported." + \
                    TRAIN_VERBOSE_DICT_INFO
                )
            for key, value in train_verbose.items():
                if not isinstance(value, TRAIN_VERBOSE_TYPE[key]):
                    raise TypeError(
                        f"train_verbose['{key}'] has wrong data type, should be {TRAIN_VERBOSE_TYPE[key]}." + \
                        TRAIN_VERBOSE_DICT_INFO
                    )
            train_verbose_.update(train_verbose)
            return train_verbose_
            
        else: raise TypeError(
            f"'train_verbose' should be of type `bool`, `int`, or `dict`, got {type(train_verbose)} instead." + \
            TRAIN_VERBOSE_DICT_INFO
        )
    
    else: raise NotImplementedError(
        f"'check_train_verbose' for 'training_type' = {training_type}"
        f" needs to be implemented."
    )


VISUALIZER_ENSEMBLES_EXAMPLE_INFO = " Example: {..., ensemble_name: ensemble, ...}"

VISUALIZER_ENSEMBLES_USAGE_INFO = \
    f" All imbalanced ensemble estimators should use the same training & validation" + \
    f" datasets and dataset names for comparable visualizations." + \
    f" Call `fit` with same 'X', 'y', 'eval_datasets'."


def _check_visualizer_ensemble_item(name, estimator) -> bool:
    if not isinstance(name, str):
        raise TypeError(
            f"Ensemble name must be `string`, got {type(name)}."
        )
    
    # Ensure estimator is an fitted sklearn/imbalanced-ensemble estimator
    # and is already fitted.
    check_is_fitted(estimator)

    if not isinstance(estimator, BaseEnsemble):
        raise TypeError(
            f"Value with name '{name}' is not an ensemble classifier instance."
        )

    if getattr(estimator, "_estimator_ensemble_type", None) == \
            "imbalanced_ensemble_classifier":
        is_imbalanced_ensemble_clf = True
    else: is_imbalanced_ensemble_clf = False

    return is_imbalanced_ensemble_clf


def get_dict_subset_by_key(dictionary:dict, subset_keys:list, exclude:bool=False):
    if exclude:
        return {k: v for k, v in dictionary.items() if k not in subset_keys}
    else: return {k: v for k, v in dictionary.items() if k in subset_keys}


def check_visualizer_ensembles(ensembles:dict, eval_datasets_:dict, eval_metrics_:dict) -> dict:

    # Check 'ensembles' parameter
    if not isinstance(ensembles, dict):
        raise TypeError(
            f"'ensembles' must be a `dict`, got {type(ensembles)}." + \
            VISUALIZER_ENSEMBLES_EXAMPLE_INFO
        )
    if len(ensembles) == 0:
        raise ValueError(
            f"'ensembles' must not be empty." + VISUALIZER_ENSEMBLES_EXAMPLE_INFO
        )
    
    # Check all key-value pairs of 'ensembles' and 
    # record names of those are not imbalanced ensemble classifier
    names_imbalanced_ensemble = []
    for name, estimator in ensembles.items():
        if _check_visualizer_ensemble_item(name, estimator):
            names_imbalanced_ensemble.append(name)
    names_sklearn_ensemble = list(set(ensembles.keys()) - set(names_imbalanced_ensemble))

    # Raise error if not all ensembles have the same n_features_
    n_features_fitted = _check_all_estimators_have_same_attribute(ensembles,
        attr_alias = ('n_features_', 'n_features_in_'))
    
    sklearn_ensembles = get_dict_subset_by_key(ensembles, names_sklearn_ensemble)
    imb_ensembles = get_dict_subset_by_key(ensembles, names_sklearn_ensemble, exclude=True)

    # Raise error if not all imbalanced ensembles have the same eval_datasets names
    if not _all_elements_equal([list(estimator.eval_datasets_.keys())
                                for estimator in imb_ensembles.values()]):
        raise ValueError(
            f"Got ensemble estimators that used inconsistent dataset names." + \
            VISUALIZER_ENSEMBLES_USAGE_INFO
        )
        
    # If eval_datasets_ is not given
    if len(eval_datasets_) == 0:
        # If all are sklearn ensemble classifier
        if len(imb_ensembles) == 0:
            raise ValueError(
                f"The 'eval_datasets' must not be empty when all "
                f" input 'ensembles' are sklearn.ensemble classifiers."
            )
        else:
            # Use imbalanced-ensemble estimators' evaluation datasets by default
            return_eval_datasets_ = copy(list(imb_ensembles.values())[0].eval_datasets_)

            # If got mixed types of ensemble classifier and eval_datasets_ is not given
            if len(sklearn_ensembles) > 0:
                warn(
                    f"the 'eval_datasets' is not specified and the input 'ensembles'"
                    f" contains sklearn.ensemble classifier, using evaluation datasets"
                    f" of other imbalanced-ensemble classifiers by default."
                )
    # If eval_datasets_ is given
    else:
        # eval_datasets_ is already validated, 
        # all data should have the same number of features
        n_features_given = list(eval_datasets_.values())[0][0].shape[1]

        # If the given data is inconsistent with the training data
        if n_features_given != n_features_fitted:
            raise ValueError(
                f"Given data in 'eval_datasets' has {n_features_given} features,"
                f" but the ensemble estimators are trained on data with"
                f" {n_features_fitted} features."
            )
        
        # Use the given evaluation datasets
        return_eval_datasets_ = copy(eval_datasets_)

    ensemble_names = list(ensembles.keys())
    dataset_names = list(return_eval_datasets_.keys())
    metric_names = list(eval_metrics_.keys())
    vis_format = {
        'n_ensembles': len(ensemble_names),
        'ensemble_names': tuple(ensemble_names),
        'n_datasets': len(dataset_names),
        'dataset_names': tuple(dataset_names),
        'n_metrics': len(metric_names),
        'metric_names': tuple(metric_names),
    }

    return ensembles, return_eval_datasets_, vis_format


def _check_all_estimators_have_same_attribute(
        ensembles:dict, attr_alias:tuple):

    has_attrs, values, not_has_attr_names = [], [], []
    for name, estimator in ensembles.items():
        recorded_flag = False
        for alias in attr_alias:
            if hasattr(estimator, alias):
                recorded_flag = True
                has_attrs.append(True)
                values.append(getattr(estimator, alias))
                break
        if not recorded_flag:
            has_attrs.append(False)
            values.append(None)
            not_has_attr_names.append(name)
    
    if not all(has_attrs):
        raise ValueError(
            f"Estimators with name {not_has_attr_names} has no"
            f" attribute {attr_alias}, check your usage." 
        )
    
    if not _all_elements_equal(values):
        raise ValueError(
            f"Got ensemble estimators that has inconsistent {attr_alias}."
            f" Make sure that the training data for all estimators"
            f" (also the evaluation data for imbalanced-ensemble estimators)"
            f" are sampled from the same task/distribution."
            )

    return values[0]


def _all_elements_equal(list_to_check:list) -> bool:
    """Private function to check whether all elements of
    list_to_check are equal."""

    # set() is not used here as some times the list 
    # elements are not hashable, e.g., strings.
    if len(list_to_check) == 1:
        return True
    return all([
        (list_to_check[i] == list_to_check[i+1])
        for i in range(len(list_to_check)-1)
        ])


PLOT_FIGSIZE_INFO = " Example: (width, height)."

def check_plot_figsize(figsize):
    if not isinstance(figsize, tuple):
        raise TypeError(
            f"'figsize' must be a tuple with 2 elements,"
            f" got {type(figsize)}." + PLOT_FIGSIZE_INFO
        )
    if len(figsize) != 2:
        raise ValueError(
            f"'figsize' must be a tuple with 2 elements,"
            f" got {len(figsize)} elements." + PLOT_FIGSIZE_INFO
        )
    for value in figsize:
        if not isinstance(value, numbers.Number):
            raise ValueError(
                f"Elements of 'figsize' must be a `int` or `float`,"
                f" got {type(value)}." + PLOT_FIGSIZE_INFO
            )
    return figsize


def check_has_diff_elements(given_set:list or set, 
                            universal_set:list or set, 
                            msg:str=""):
    diff_set = set(given_set) - set(universal_set)
    if len(diff_set) > 0:
        raise ValueError(
            msg % {"diff_set": diff_set}
        )


def check_type(param, param_name:str, typ, typ_name:str=None):
    if not isinstance(param, typ):
        typ_name = str(typ) if typ_name is None else typ_name
        raise ValueError(
            f"'{param_name}' should be of type `{typ_name}`,"
            f" got {type(param)}."
        )
    return param

    
def check_pred_proba(y_pred_proba, n_samples, n_classes, dtype=None):
    """Private function for validating y_pred_proba"""
    if dtype is not None and dtype not in [np.float32, np.float64]:
        dtype = np.float64

    if dtype is None:
        dtype = [np.float64, np.float32]
    y_pred_proba = check_array(
        y_pred_proba, accept_sparse=False, ensure_2d=False, dtype=dtype,
        order="C"
    )
    if y_pred_proba.ndim != 2:
        raise ValueError("Predicted probabilites must be 2D array")

    if y_pred_proba.shape != (n_samples, n_classes):
        raise ValueError("y_pred_proba.shape == {}, expected {}!"
                        .format(y_pred_proba.shape, (n_samples, n_classes)))
    return y_pred_proba



"""
------------------------------------------------------------------------------
The `self_paced_ensemble.canonical_ensemble` module implement 5 ensemble 
learning algorithms for imbalanced classification, including:
'SMOTEBaggingClassifier', 'SMOTEBoostClassifier', 'RUSBoostClassifier', 
'UnderBaggingClassifier', and 'BalanceCascadeClassifier'.

Note: methods in this module are now included in the `imbalanced-ensemble`.
Please refer to https://imbalanced-ensemble.readthedocs.io/ for more details.
------------------------------------------------------------------------------
"""

from .canonical_ensemble import (
    SMOTEBaggingClassifier,
    SMOTEBoostClassifier,
    RUSBoostClassifier,
    UnderBaggingClassifier,
    BalanceCascadeClassifier,
)

__all__ = [
    "SMOTEBaggingClassifier",
    "SMOTEBoostClassifier",
    "RUSBoostClassifier",
    "UnderBaggingClassifier",
    "BalanceCascadeClassifier",
]



# -*- coding: utf-8 -*-
"""
A self-paced Ensemble (SPE) Classifier for binary class-imbalanced learning.
    
Self-paced Ensemble (SPE) is an ensemble learning framework for massive highly 
imbalanced classification. It is an easy-to-use solution to class-imbalanced 
problems, features outstanding computing efficiency, good performance, and wide 
compatibility with different learning models.
"""

# Created on Tue May 14 14:32:27 2019
# @author: ZhiningLiu1998
# mailto:  zhining.liu@outlook.com

# %%

from collections import Counter
import numpy as np
from joblib import dump


from .base import BaseImbalancedEnsemble, MAX_INT
from ._self_paced_under_sampler import SelfPacedUnderSampler
from ..utils._validation import _deprecate_positional_args
from ..utils._validation_data import check_eval_datasets
from ..utils._validation_param import (check_target_label_and_n_target_samples, 
                                       check_balancing_schedule, 
                                       check_train_verbose, 
                                       check_eval_metrics,)

# # For local test
# import sys
# sys.path.append("..")
# from self_paced_ensemble.base import BaseImbalancedEnsemble, MAX_INT
# from self_paced_ensemble._self_paced_under_sampler import SelfPacedUnderSampler
# from utils._validation import _deprecate_positional_args
# from utils._validation_data import check_eval_datasets
# from utils._validation_param import (check_target_label_and_n_target_samples, 
#                                      check_balancing_schedule, 
#                                      check_train_verbose, 
#                                      check_eval_metrics,)

# %%

# Properties
_method_name = 'SelfPacedEnsembleClassifier'
_sampler_class = SelfPacedUnderSampler

_solution_type = 'resampling'
_sampling_type = 'under-sampling'
_ensemble_type = 'general'
_training_type = 'iterative'

_properties = {
    'solution_type': _solution_type,
    'sampling_type': _sampling_type,
    'ensemble_type': _ensemble_type,
    'training_type': _training_type,
}

class SelfPacedEnsembleClassifier(BaseImbalancedEnsemble):
    """A self-paced ensemble (SPE) Classifier for class-imbalanced learning.
    
    Self-paced Ensemble (SPE) [1]_ is an ensemble learning framework for massive highly 
    imbalanced classification. It is an easy-to-use solution to class-imbalanced 
    problems, features outstanding computing efficiency, good performance, and wide 
    compatibility with different learning models.

    This implementation extends SPE to support multi-class classification.

    Parameters
    ----------
    base_estimator : estimator object, default=None
        The base estimator to fit on self-paced under-sampled subsets 
        of the dataset. Support for sample weighting is NOT required, 
        but need proper ``classes_`` and ``n_classes_`` attributes. 
        If ``None``, then the base estimator is ``DecisionTreeClassifier()``.

    n_estimators : int, default=50
        The number of base estimators in the ensemble.

    k_bins : int, default=5
        The number of hardness bins that were used to approximate 
        hardness distribution. It is recommended to set it to 5. 
        One can try a larger value when the smallest class in the 
        data set has a sufficient number (say, > 1000) of samples.

    soft_resample_flag : bool, default=False
        Whether to use weighted sampling to perform soft self-paced 
        under-sampling, rather than explicitly cut samples into 
        ``k``-bins and perform hard sampling.

    replacement : bool, default=True
        Whether samples are drawn with replacement. If ``False`` 
        and ``soft_resample_flag = False``, may raise an error when 
        a bin has insufficient number of data samples for resampling.

    estimator_params : list of str, default=tuple()
        The list of attributes to use as parameters when instantiating a
        new base estimator. If none are given, default parameters are used.

    {n_jobs}

    {random_state}

    verbose : int, default=0
        Controls the verbosity when predicting.

    Attributes
    ----------
    base_estimator : estimator
        The base estimator from which the ensemble is grown.

    base_sampler_ : SelfPacedUnderSampler
        The base sampler.

    estimators_ : list of estimator
        The collection of fitted base estimators.

    samplers_ : list of SelfPacedUnderSampler
        The collection of fitted samplers.

    classes_ : ndarray of shape (n_classes,)
        The classes labels.

    n_classes_ : int
        The number of classes.

    feature_importances_ : ndarray of shape (n_features,)
        The feature importances if supported by the ``base_estimator``.
    
    estimators_n_training_samples_ : list of ints
        The number of training samples for each fitted 
        base estimators.

    See Also
    --------
    BalanceCascadeClassifier : Ensemble with cascade dynamic under-sampling.

    EasyEnsembleClassifier : Bag of balanced boosted learners.

    RUSBoostClassifier : Random under-sampling integrated in AdaBoost.

    Notes
    -----
    See :ref:`sphx_glr_auto_examples_basic_plot_basic_example.py` for an example.

    References
    ----------
    .. [1] Liu, Z., Cao, W., Gao, Z., Bian, J., Chen, H., Chang, Y., & Liu, T. Y. 
       "Self-paced ensemble for highly imbalanced massive data classification." 
       2020 IEEE 36th International Conference on Data Engineering (ICDE). 
       IEEE, 2010: 841-852.

    Examples
    --------
    {example}
    """
    
    def __init__(self, 
        base_estimator=None, 
        n_estimators:int=50, 
        k_bins:int=5, 
        soft_resample_flag:bool=False, 
        replacement:bool=True, 
        estimator_params=tuple(),
        n_jobs=None,
        random_state=None,
        verbose=0,):
    
        super(SelfPacedEnsembleClassifier, self).__init__( 
            base_estimator=base_estimator, 
            n_estimators=n_estimators,
            estimator_params=estimator_params,
            random_state=random_state,
            n_jobs=n_jobs, 
            verbose=verbose)

        self.__name__ = _method_name
        self.base_sampler = _sampler_class()
        self._sampling_type = _sampling_type
        self._sampler_class = _sampler_class
        self._properties = _properties

        self.k_bins = k_bins
        self.soft_resample_flag = soft_resample_flag
        self.replacement = replacement
    

    def fit(self, X, y, *, sample_weight=None, **kwargs):
        """Build a SPE classifier from the training set (X, y).

        Parameters
        ----------
        X : {array-like, sparse matrix} of shape (n_samples, n_features)
            The training input samples. Sparse matrix can be CSC, CSR, COO,
            DOK, or LIL. DOK and LIL are converted to CSR.

        y : array-like of shape (n_samples,)
            The target values (class labels).

        sample_weight : array-like of shape (n_samples,), default=None
            Sample weights. If None, the sample weights are initialized to
            ``1 / n_samples``.
        
        target_label : int, default=None
            Specify the class targeted by the under-sampling. 
            All other classes that have more samples than the target class will 
            be considered as majority classes. They will be under-sampled until 
            the number of samples is equalized. The remaining minority classes 
            (if any) will stay unchanged.
        
        n_target_samples : int or dict, default=None
            Specify the desired number of samples (of each class) after the 
            under-sampling. 

            - If ``int``, all classes that have more than the ``n_target_samples`` 
              samples will be under-sampled until the number of samples is equalized.
            - If ``dict``, the keys correspond to the targeted classes. The values 
              correspond to the desired number of samples for each targeted class.
        
        balancing_schedule : str, or callable, default='uniform'
            Scheduler that controls how to sample the data set during the ensemble 
            training process.

            - If ``str``, using the predefined balancing schedule.
              Possible choices are:

                - ``'uniform'``: resample to target distribution for all base estimators;
                - ``'progressive'``: The resample class distributions are progressive 
                  interpolation between the original and the target class distribution.
                  Example: For a class :math:`c`, say the number of samples is :math:`N_{c}` 
                  and the target number of samples is :math:`N'_{c}`. Suppose that we are 
                  training the :math:`t`-th base estimator of a :math:`T`-estimator ensemble, then 
                  we expect to get :math:`(1-\frac{t}{T}) \cdot N_{c} + \frac{t}{T} \cdot N'_{c}` 
                  samples after resampling;

            - If callable, function takes 4 positional arguments with order (``'origin_distr'``: 
              ``dict``, ``'target_distr'``: ``dict``, ``'i_estimator'``: ``int``, ``'total_estimator'``: 
              ``int``), and returns a ``'result_distr'``: ``dict``. For all parameters of type ``dict``, 
              the keys of type ``int`` correspond to the targeted classes, and the values of type ``str`` 
              correspond to the (desired) number of samples for each class.
        
        eval_datasets : dict, default=None
            Dataset(s) used for evaluation during the ensemble training process.
            The keys should be strings corresponding to evaluation datasets' names. 
            The values should be tuples corresponding to the input samples and target
            values. 
            
            Example: ``eval_datasets = {'valid' : (X_valid, y_valid)}``
        
        eval_metrics : dict, default=None
            Metric(s) used for evaluation during the ensemble training process.

            - If ``None``, use 3 default metrics:

                - ``'acc'``: 
                  ``sklearn.metrics.accuracy_score()``
                - ``'balanced_acc'``: 
                  ``sklearn.metrics.balanced_accuracy_score()``
                - ``'weighted_f1'``: 
                  ``sklearn.metrics.f1_score(average='weighted')``

            - If ``dict``, the keys should be strings corresponding to evaluation 
              metrics' names. The values should be tuples corresponding to the metric 
              function (``callable``) and additional kwargs (``dict``).

                - The metric function should at least take 2 named/keyword arguments, 
                  ``y_true`` and one of [``y_pred``, ``y_score``], and returns a float
                  as the evaluation score. Keyword arguments:

                  - ``y_true``, 1d-array of shape (n_samples,), true labels or binary 
                    label indicators corresponds to ground truth (correct) labels.
                  - When using ``y_pred``, input will be 1d-array of shape (n_samples,) 
                    corresponds to predicted labels, as returned by a classifier.
                  - When using ``y_score``, input will be 2d-array of shape (n_samples, 
                    n_classes,) corresponds to probability estimates provided by the 
                    predict_proba method. In addition, the order of the class scores 
                    must correspond to the order of ``labels``, if provided in the metric 
                    function, or else to the numerical or lexicographical order of the 
                    labels in ``y_true``.
                
                - The metric additional kwargs should be a dictionary that specifies 
                  the additional arguments that need to be passed into the metric function. 
            
            Example: ``{'weighted_f1': (sklearn.metrics.f1_score, {'average': 'weighted'})}``
        
        train_verbose : bool, int or dict, default=False
            Controls the verbosity during ensemble training/fitting.

            - If ``bool``: ``False`` means disable training verbose. ``True`` means 
              print training information to sys.stdout use default setting:
              
                - ``'granularity'``        : ``int(n_estimators/10)``
                - ``'print_distribution'`` : ``True``
                - ``'print_metrics'``      : ``True``

            - If ``int``, print information per ``train_verbose`` rounds.

            - If ``dict``, control the detailed training verbose settings. They are:

                - ``'granularity'``: corresponding value should be ``int``, the training
                  information will be printed per ``granularity`` rounds.
                - ``'print_distribution'``: corresponding value should be ``bool``, 
                  whether to print the data class distribution 
                  after resampling. Will be ignored if the 
                  ensemble training does not perform resampling.
                - ``'print_metrics'``: corresponding value should be ``bool``, 
                  whether to print the latest performance score.
                  The performance will be evaluated on the training 
                  data and all given evaluation datasets with the 
                  specified metrics.
              
            .. warning::
                Setting a small ``'granularity'`` value with ``'print_metrics'`` enabled 
                can be costly when the training/evaluation data is large or the metric 
                scores are hard to compute. Normally, one can set ``'granularity'`` to 
                ``n_estimators/10`` (this is used by default).

        Returns
        -------
        self : object
        """
        return super().fit(
            X, y, sample_weight=sample_weight, **kwargs
        )


    @_deprecate_positional_args
    def _fit(self, X, y, 
            *,
            sample_weight=None, 
            target_label:int=None, 
            n_target_samples:int or dict=None, 
            balancing_schedule:str or function='uniform',
            eval_datasets:dict=None,
            eval_metrics:dict=None,
            train_verbose:bool or int or dict=False,
            ):
        
        # X, y, sample_weight, base_estimators_ (default=DecisionTreeClassifier),
        # n_estimators, random_state, sample_weight are already validated in super.fit()
        random_state, n_estimators, replacement, k_bins, soft_resample_flag, classes_ = \
            self.random_state, self.n_estimators, self.replacement, self.k_bins, \
            self.soft_resample_flag, self.classes_

        # Check evaluation data
        check_x_y_args = self.check_x_y_args
        self.eval_datasets_ = check_eval_datasets(eval_datasets, X, y, **check_x_y_args)

        # Check target sample strategy
        origin_distr_ = dict(Counter(y))
        target_label_, target_distr_ = \
            check_target_label_and_n_target_samples(
                y, target_label, n_target_samples, self._sampling_type)
        self.origin_distr_, self.target_label_, self.target_distr_ = \
            origin_distr_, target_label_, target_distr_
        
        # Check balancing schedule
        balancing_schedule_ = check_balancing_schedule(balancing_schedule)
        self.balancing_schedule_ = balancing_schedule_

        # Check evaluation metrics
        self.eval_metrics_ = check_eval_metrics(eval_metrics)
        
        # Check training train_verbose format
        self.train_verbose_ = check_train_verbose(
            train_verbose, self.n_estimators, **self._properties)
        
        # Set training verbose format
        self._init_training_log_format()

        # Clear any previous fit results.
        self.estimators_ = []
        self.estimators_features_ = []
        self.estimators_n_training_samples_ = np.zeros(n_estimators, dtype=int)
        self.samplers_ = []
        
        # Genrate random seeds array
        seeds = random_state.randint(MAX_INT, size=n_estimators)
        self._seeds = seeds

        # Check if sample_weight is specified
        specified_sample_weight = (sample_weight is not None)

        for i_iter in range(n_estimators):

            current_iter_distr = balancing_schedule_(
                origin_distr=origin_distr_,
                target_distr=target_distr_,
                i_estimator=i_iter,
                total_estimator=n_estimators,
            )

            sampler = self._make_sampler(
                append=True,
                random_state=seeds[i_iter],
                sampling_strategy=current_iter_distr,
                k_bins=k_bins,
                soft_resample_flag=soft_resample_flag, 
                replacement=replacement, 
            )
            
            # update self.y_pred_proba_latest
            self._update_cached_prediction_probabilities(i_iter, X)
            
            # compute alpha
            alpha = np.tan(np.pi*0.5*(i_iter/(max(n_estimators-1, 1))))

            # Perform self-paced under-sampling
            resample_out = sampler.fit_resample(X, y, 
                    y_pred_proba=self.y_pred_proba_latest,
                    alpha=alpha,
                    classes_=classes_,
                    sample_weight=sample_weight)

            # Train a new base estimator on resampled data 
            # and add it into self.estimators_
            estimator = self._make_estimator(append=True, random_state=seeds[i_iter])
            if specified_sample_weight:
                (X_resampled, y_resampled, sample_weight_resampled) = resample_out
                estimator.fit(X_resampled, y_resampled, sample_weight=sample_weight_resampled)
            else: 
                (X_resampled, y_resampled) = resample_out
                estimator.fit(X_resampled, y_resampled)

            self.estimators_features_.append(self.features_)
            self.estimators_n_training_samples_[i_iter] = y_resampled.shape[0]

            # Print training infomation to console.
            self._training_log_to_console(i_iter, y_resampled)
        
        return self

    
    def _update_cached_prediction_probabilities(self, i_iter, X):
        """Private function that maintains a latest prediction probabilities of the training
         data during ensemble training. Must be called in each iteration before fit the 
         base_estimator."""

        if i_iter == 0:
            self.y_pred_proba_latest = np.zeros((self._n_samples, self.n_classes_), 
                                                dtype=np.float64)
        else:
            y_pred_proba_latest = self.y_pred_proba_latest
            y_pred_proba_new = self.estimators_[-1].predict_proba(X)
            self.y_pred_proba_latest = (y_pred_proba_latest * i_iter + y_pred_proba_new) / (i_iter+1)
        return
    
    def save(self, filename:str):
        """Save the model to a file."""
        dump(self, filename=filename)

# %%

if __name__ == '__main__':
    from collections import Counter
    from copy import copy
    from sklearn.tree import DecisionTreeClassifier
    from sklearn.datasets import make_classification
    from sklearn.model_selection import train_test_split
    from sklearn.metrics import accuracy_score, balanced_accuracy_score, f1_score
    
    # X, y = make_classification(n_classes=2, class_sep=2, # 2-class
    #     weights=[0.1, 0.9], n_informative=3, n_redundant=1, flip_y=0,
    #     n_features=20, n_clusters_per_class=1, n_samples=1000, random_state=10)
    X, y = make_classification(n_classes=3, class_sep=2, # 3-class
        weights=[0.1, 0.3, 0.6], n_informative=3, n_redundant=1, flip_y=0,
        n_features=20, n_clusters_per_class=1, n_samples=2000, random_state=10)

    X_train, X_valid, y_train, y_valid = train_test_split(X, y, test_size=0.5, random_state=42)

    origin_distr = dict(Counter(y_train)) # {2: 600, 1: 300, 0: 100}
    print('Original training dataset shape %s' % origin_distr)

    target_distr = {2: 200, 1: 100, 0: 100}

    init_kwargs_default = {
        'base_estimator': None,
        # 'base_estimator': DecisionTreeClassifier(max_depth=10),
        'n_estimators': 100,
        'k_bins': 5,
        'soft_resample_flag': False,
        'replacement': True,
        'estimator_params': tuple(),
        'n_jobs': None,
        'random_state': 42,
        # 'random_state': None,
        'verbose': 0,
    }

    fit_kwargs_default = {
        'X': X_train,
        'y': y_train,
        'sample_weight': None,
        'target_label': None,
        'n_target_samples': None,
        # 'n_target_samples': target_distr,
        'balancing_schedule': 'uniform',
        'eval_datasets': {'valid': (X_valid, y_valid)},
        'eval_metrics': {
            'acc': (accuracy_score, {}),
            'balanced_acc': (balanced_accuracy_score, {}),
            'weighted_f1': (f1_score, {'average':'weighted'}),},
        'train_verbose': {
            'granularity': 10,
            'print_distribution': True,
            'print_metrics': True,},
    }

    ensembles = {}

    init_kwargs, fit_kwargs = copy(init_kwargs_default), copy(fit_kwargs_default)
    spe = SelfPacedEnsembleClassifier(**init_kwargs).fit(**fit_kwargs)
    ensembles['spe'] = spe

    init_kwargs, fit_kwargs = copy(init_kwargs_default), copy(fit_kwargs_default)
    fit_kwargs.update({
        'balancing_schedule': 'progressive'
    })
    spe_prog = SelfPacedEnsembleClassifier(**init_kwargs).fit(**fit_kwargs)
    ensembles['spe_prog'] = spe_prog

    init_kwargs, fit_kwargs = copy(init_kwargs_default), copy(fit_kwargs_default)
    init_kwargs.update({
        'soft_resample_flag': True,
        'replacement': False,
    })
    spe_soft = SelfPacedEnsembleClassifier(**init_kwargs).fit(**fit_kwargs)
    ensembles['spe_soft'] = spe_soft


    # %%
    from imbens.visualizer import ImbalancedEnsembleVisualizer

    visualizer = ImbalancedEnsembleVisualizer(
        eval_datasets = None,
        eval_metrics = None,
    ).fit(
        ensembles = ensembles,
        granularity = 5,
    )
    fig, axes = visualizer.performance_lineplot(
        on_ensembles=None,
        on_datasets=None,
        split_by=[],
        n_samples_as_x_axis=False,
        sub_figsize=(4, 3.3),
        sup_title=True,
        alpha=0.8,
    )
    fig, axes = visualizer.confusion_matrix_heatmap(
        on_ensembles=None,
        on_datasets=None,
        sub_figsize=(4, 3.3),
    )

    # %%



"""Utilities for data visualization."""

# Authors: Zhining Liu <zhining.liu@outlook.com>
# License: MIT

import pandas as pd
import numpy as np

from copy import copy
from collections import Counter
import matplotlib.pyplot as plt
import seaborn as sns

from sklearn.decomposition import KernelPCA

DEFAULT_VIS_KWARGS = {
    'cmap': plt.cm.rainbow,
    'edgecolor': 'black',
    'alpha': 0.6,
}

def set_ax_border(ax, border_color='black', border_width=2):
    for _, spine in ax.spines.items():
        spine.set_color(border_color)
        spine.set_linewidth(border_width)

    return ax

def plot_scatter(X, y, ax=None, weights=None, title='',
                 projection=None, vis_params=None):
    if ax is None:
        ax = plt.axes()
    X_vis = projection.transform(X) if X.shape[1] > 2 else X
    title += ' (2D projection by {})'.format(
        str(projection.__class__).split('.')[-1][:-2]
    )
    size = 50 if weights is None else weights
    if np.unique(y).shape[0] > 2:
        vis_params['palette'] = plt.cm.rainbow
    sns.scatterplot(x=X_vis[:, 0], y=X_vis[:, 1],
        hue=y, style=y, s=size, **vis_params, legend='full', ax=ax)

    ax.set_title(title)
    ax = set_ax_border(ax, border_color='black', border_width=2)
    ax.grid(color='black', linestyle='-.', alpha=0.5)

    return ax

def plot_class_distribution(y, ax=None, title='',
                            sort_values=False, plot_average=True):
    count = pd.DataFrame(list(Counter(y).items()),
                         columns=['Class', 'Frequency'])
    if sort_values:
        count = count.sort_values(by='Frequency', ascending=False)
    if ax is None:
        ax = plt.axes()
    count.plot.bar(x='Class', y='Frequency', title=title, ax=ax)

    ax.set_title(title)
    ax = set_ax_border(ax, border_color='black', border_width=2)
    ax.grid(color='black', linestyle='-.', alpha=0.5, axis='y')

    if plot_average:
        ax.axhline(y=count['Frequency'].mean(),ls="dashdot",c="red")
        xlim_min, xlim_max, ylim_min, ylim_max = ax.axis()
        ax.text(
            x=xlim_min+(xlim_max-xlim_min)*0.82,
            y=count['Frequency'].mean()+(ylim_max-ylim_min)*0.03,
            c="red",s='Average')

    return ax

def plot_2Dprojection_and_cardinality(X, y, figsize=(10, 4), vis_params=None,
                                     projection=None, weights=None, plot_average=True,
                                     title1='Dataset', title2='Class Distribution'):

    if vis_params is None:
        vis_params = copy(DEFAULT_VIS_KWARGS)

    fig, (ax1, ax2) = plt.subplots(1, 2, figsize=figsize)
    if projection == None:
        projection = KernelPCA(n_components=2).fit(X, y)
    ax1 = plot_scatter(X, y, ax=ax1, weights=weights, title=title1,
                    projection=projection, vis_params=vis_params)
    ax2 = plot_class_distribution(y, ax=ax2, title=title2,
                    sort_values=True, plot_average=plot_average)
    plt.tight_layout()
    return fig


"""
``self-paced ensemble`` is a python-based ensemble learning framework for 
dealing with binary class-imbalanced classification problems in machine learning.
"""
# License: MIT

__version__ = '0.1.7'


# -*- coding: utf-8 -*-
"""Utilities."""

# Authors: Zhining Liu <zhining.liu@outlook.com>
# License: MIT

from joblib import dump, load
from sklearn.metrics import (
    precision_recall_curve, 
    average_precision_score, 
    matthews_corrcoef,
    )
from sklearn import datasets
from sklearn.model_selection import train_test_split
import numpy as np

def load_covtype_dataset(subset=0.1, test_size=0.2, random_state=None):
    '''Load & Split training/test covtype dataset.'''
    print ('\nDataset used: \t\tForest covertypes from UCI ({:.1%} random subset)'.format(subset))
    X, y = datasets.fetch_covtype(return_X_y=True)
    y = make_binary_classification_target(y, 7, verbose=True)
    X, y = imbalance_random_subset(
        X, y, size=subset, random_state=random_state)
    X_train, X_test, y_train, y_test = imbalance_train_test_split(
        X, y, test_size=test_size, random_state=random_state)
    return X_train, X_test, y_train, y_test

def make_binary_classification_target(y, pos_label, verbose=False):
    '''Turn multi-class targets into binary classification targets.'''
    pos_idx = (y==pos_label)
    y[pos_idx] = 1
    y[~pos_idx] = 0
    if verbose:
        print ('Positive target:\t{}'.format(pos_label))
        print ('Imbalance ratio:\t{:.3f}'.format((y==0).sum()/(y==1).sum()))
    return y

def imbalance_train_test_split(X, y, test_size, random_state=None):
    '''Train/Test split that guarantee same class distribution between split datasets.'''
    X_maj = X[y==0]; y_maj = y[y==0]
    X_min = X[y==1]; y_min = y[y==1]
    X_train_maj, X_test_maj, y_train_maj, y_test_maj = train_test_split(
        X_maj, y_maj, test_size=test_size, random_state=random_state)
    X_train_min, X_test_min, y_train_min, y_test_min = train_test_split(
        X_min, y_min, test_size=test_size, random_state=random_state)
    X_train = np.concatenate([X_train_maj, X_train_min])
    X_test = np.concatenate([X_test_maj, X_test_min])
    y_train = np.concatenate([y_train_maj, y_train_min])
    y_test = np.concatenate([y_test_maj, y_test_min])
    return  X_train, X_test, y_train, y_test

def imbalance_random_subset(X, y, size, random_state=None):
    '''Get random subset while guarantee same class distribution.'''
    _, X, _, y = imbalance_train_test_split(X, y, 
        test_size=size, random_state=random_state)
    return X, y

def auc_prc(label, y_pred):
    '''Compute AUCPRC score.'''
    return average_precision_score(label, y_pred)

def f1_optim(label, y_pred):
    '''Compute optimal F1 score.'''
    y_pred = y_pred.copy()
    prec, reca, _ = precision_recall_curve(label, y_pred)
    f1s = 2 * (prec * reca) / (prec + reca)
    return max(f1s)

def gm_optim(label, y_pred):
    '''Compute optimal G-mean score.'''
    y_pred = y_pred.copy()
    prec, reca, _ = precision_recall_curve(label, y_pred)
    gms = np.power((prec*reca), 0.5)
    return max(gms)

def mcc_optim(label, y_pred):
    '''Compute optimal MCC score.'''
    mccs = []
    for t in range(100):
        y_pred_b = y_pred.copy()
        y_pred_b[y_pred_b < 0+t*0.01] = 0
        y_pred_b[y_pred_b >= 0+t*0.01] = 1
        mcc = matthews_corrcoef(label, y_pred_b)
        mccs.append(mcc)
    return max(mccs)

def save_model(model, filename:str):
    """Save model to file."""
    dump(model, filename=filename)

def load_model(filename:str):
    """Load model from file."""
    return load(filename)


"""Base class for sampling."""

# Authors: Zhining Liu <zhining.liu@outlook.com>
# License: MIT

# %%

from abc import ABCMeta, abstractmethod

import numpy as np

from sklearn.base import BaseEstimator
from sklearn.utils.validation import _check_sample_weight
from sklearn.preprocessing import label_binarize
from sklearn.utils.multiclass import check_classification_targets


from ..utils._validation import (ArraysTransformer, 
                                 _deprecate_positional_args, 
                                 check_sampling_strategy, 
                                 check_target_type)

# # For local test
# import sys
# sys.path.append("..")
# from utils._validation import (ArraysTransformer, 
#                                _deprecate_positional_args, 
#                                check_sampling_strategy, 
#                                check_target_type)


class SamplerMixin(BaseEstimator, metaclass=ABCMeta):
    """Mixin class for samplers with abstract method.

    Warning: This class should not be used directly. Use the derive classes
    instead.
    """

    _estimator_type = "sampler"

    def fit(self, X, y):
        """Check inputs and statistics of the sampler.

        You should use ``fit_resample`` in all cases.

        Parameters
        ----------
        X : {array-like, dataframe, sparse matrix} of shape \
                (n_samples, n_features)
            Data array.

        y : array-like of shape (n_samples,)
            Target array.

        Returns
        -------
        self : object
            Return the instance itself.
        """
        X, y, _ = self._check_X_y(X, y)
        self.sampling_strategy_ = check_sampling_strategy(
            self.sampling_strategy, y, self._sampling_type
        )
        return self

    @_deprecate_positional_args
    def fit_resample(self, X, y, *, sample_weight=None, **kwargs):
        """Resample the dataset.

        Parameters
        ----------
        X : {array-like, dataframe, sparse matrix} of shape \
                (n_samples, n_features)
            Matrix containing the data which have to be sampled.

        y : array-like of shape (n_samples,)
            Corresponding label for each sample in X.

        sample_weight : array-like of shape (n_samples,), default=None
            Corresponding weight for each sample in X. 

            - If ``None``, perform normal resampling and return 
              ``(X_resampled, y_resampled)``.
            - If array-like, the given ``sample_weight`` will be resampled 
              along with ``X`` and ``y``, and the *resampled* sample weights 
              will be added to returns. The function will return
              ``(X_resampled, y_resampled, sample_weight_resampled)``.

        Returns
        -------
        X_resampled : {array-like, dataframe, sparse matrix} of shape \
                (n_samples_new, n_features)
            The array containing the resampled data.

        y_resampled : array-like of shape (n_samples_new,)
            The corresponding label of `X_resampled`.
        
        sample_weight_resampled : array-like of shape (n_samples_new,), default=None
            The corresponding weight of `X_resampled`.
            *Only will be returned if input sample_weight is not* ``None``.
        """
        check_classification_targets(y)
        arrays_transformer = ArraysTransformer(X, y)
        X, y, binarize_y = self._check_X_y(X, y)

        self.sampling_strategy_ = check_sampling_strategy(
            self.sampling_strategy, y, self._sampling_type
        )

        if sample_weight is None:
            output = self._fit_resample(X, y, **kwargs)
        else:
            try:
                sample_weight = _check_sample_weight(sample_weight, X, dtype=np.float64)
            except Exception as e:
                e_args = list(e.args)
                e_args[0] += \
                    f"\n'sample_weight' should be an array-like of shape (n_samples,)," + \
                    f" got {type(sample_weight)}, please check your usage."
                e.args = tuple(e_args)
                raise e
            else:
                output = self._fit_resample(X, y, sample_weight=sample_weight, **kwargs)

        y_ = label_binarize(output[1], classes=np.unique(y)) if binarize_y else output[1]

        X_, y_ = arrays_transformer.transform(output[0], y_)
        return (X_, y_) if len(output) == 2 else output

    @abstractmethod
    def _fit_resample(self, X, y, sample_weight=None):
        """Base method defined in each sampler to defined the sampling
        strategy.

        Parameters
        ----------
        X : {array-like, sparse matrix} of shape (n_samples, n_features)
            Matrix containing the data which have to be sampled.

        y : array-like of shape (n_samples,)
            Corresponding label for each sample in X.

        Returns
        -------
        X_resampled : {ndarray, sparse matrix} of shape \
                (n_samples_new, n_features)
            The array containing the resampled data.

        y_resampled : ndarray of shape (n_samples_new,)
            The corresponding label of `X_resampled`.

        """
        pass


class BaseSampler(SamplerMixin):
    """Base class for sampling algorithms.

    Warning: This class should not be used directly. Use the derive classes
    instead.
    """

    def __init__(self, sampling_strategy="auto"):
        self.sampling_strategy = sampling_strategy

    def _check_X_y(self, X, y, accept_sparse=None):
        if accept_sparse is None:
            accept_sparse = ["csr", "csc"]
        y, binarize_y = check_target_type(y, indicate_one_vs_all=True)
        X, y = self._validate_data(X, y, reset=True, accept_sparse=accept_sparse)
        return X, y, binarize_y

    def _more_tags(self):
        return {"X_types": ["2darray", "sparse", "dataframe"]}


def _identity(X, y):
    return X, y


class FunctionSampler(BaseSampler):
    """Construct a sampler from calling an arbitrary callable.

    Parameters
    ----------
    func : callable, default=None
        The callable to use for the transformation. This will be passed the
        same arguments as transform, with args and kwargs forwarded. If func is
        None, then func will be the identity function.

    accept_sparse : bool, default=True
        Whether sparse input are supported. By default, sparse inputs are
        supported.

    kw_args : dict, default=None
        The keyword argument expected by ``func``.

    validate : bool, default=True
        Whether or not to bypass the validation of ``X`` and ``y``. Turning-off
        validation allows to use the ``FunctionSampler`` with any type of
        data.

    See Also
    --------

    sklearn.preprocessing.FunctionTransfomer : Stateless transformer.
    """

    _sampling_type = "bypass"

    @_deprecate_positional_args
    def __init__(self, *, func=None, accept_sparse=True, kw_args=None, validate=True):
        super().__init__()
        self.func = func
        self.accept_sparse = accept_sparse
        self.kw_args = kw_args
        self.validate = validate

    def fit(self, X, y):
        """Check inputs and statistics of the sampler.

        You should use ``fit_resample`` in all cases.

        Parameters
        ----------
        X : {array-like, dataframe, sparse matrix} of shape \
                (n_samples, n_features)
            Data array.

        y : array-like of shape (n_samples,)
            Target array.

        Returns
        -------
        self : object
            Return the instance itself.
        """
        # we need to overwrite SamplerMixin.fit to bypass the validation
        if self.validate:
            check_classification_targets(y)
            X, y, _ = self._check_X_y(X, y, accept_sparse=self.accept_sparse)

        self.sampling_strategy_ = check_sampling_strategy(
            self.sampling_strategy, y, self._sampling_type
        )

        return self

    def fit_resample(self, X, y):
        """Resample the dataset.

        Parameters
        ----------
        X : {array-like, sparse matrix} of shape (n_samples, n_features)
            Matrix containing the data which have to be sampled.

        y : array-like of shape (n_samples,)
            Corresponding label for each sample in X.

        Returns
        -------
        X_resampled : {array-like, sparse matrix} of shape \
                (n_samples_new, n_features)
            The array containing the resampled data.

        y_resampled : array-like of shape (n_samples_new,)
            The corresponding label of `X_resampled`.
        """
        arrays_transformer = ArraysTransformer(X, y)

        if self.validate:
            check_classification_targets(y)
            X, y, binarize_y = self._check_X_y(X, y, accept_sparse=self.accept_sparse)

        self.sampling_strategy_ = check_sampling_strategy(
            self.sampling_strategy, y, self._sampling_type
        )

        output = self._fit_resample(X, y)

        if self.validate:

            y_ = label_binarize(output[1], classes=np.unique(y)) if binarize_y else output[1]
            X_, y_ = arrays_transformer.transform(output[0], y_)
            return (X_, y_) if len(output) == 2 else (X_, y_, output[2])

        return output

    def _fit_resample(self, X, y):
        func = _identity if self.func is None else self.func
        output = func(X, y, **(self.kw_args if self.kw_args else {}))
        return output

# %%



"""Mixin and base class for ensemble classifiers.
"""

# Authors: Zhining Liu <zhining.liu@outlook.com>
# License: MIT


# %%

from abc import ABCMeta, abstractmethod

import numpy as np
from collections import Counter
from joblib import Parallel
import numpy as np
from inspect import signature

from sklearn.base import ClassifierMixin, clone
from sklearn.ensemble import BaseEnsemble
from sklearn.ensemble._base import _set_random_states
from sklearn.ensemble._base import _partition_estimators
from sklearn.ensemble._bagging import _parallel_predict_proba
from sklearn.tree import DecisionTreeClassifier
from sklearn.utils import check_random_state
from sklearn.utils import check_array
from sklearn.utils.fixes import delayed
from sklearn.utils.multiclass import check_classification_targets
from sklearn.utils.validation import (_check_sample_weight, 
                                      check_random_state, 
                                      check_is_fitted, 
                                      column_or_1d, 
                                      check_array,
                                      has_fit_parameter,)


from ..utils._validation import _deprecate_positional_args

# # For local test
# import sys
# sys.path.append("..")
# from utils._validation import _deprecate_positional_args

TRAINING_LOG_HEAD_TITLES = {
    'iter': '#Estimators',
    'class_distr': 'Class Distribution',
    'datasets': 'Datasets',
    'metrics': 'Metrics',
}

TRAINING_TYPES = ('iterative', 'parallel')

MAX_INT = np.iinfo(np.int32).max

# %%

def sort_dict_by_key(d):
    """Sort a dict by key, return sorted dict."""
    return dict(sorted(d.items(), key=lambda k: k[0]))


class ImbalancedEnsembleClassifierMixin(ClassifierMixin):
    """Mixin class for all ensemble classifiers in imbalanced-ensemble.

    This class is essential for a derived class to be identified by the
    sklearn and imbalanced-ensemble package. Additionally, it provides 
    several utilities for formatting training logs of imbalanced-ensemble 
    classifiers.
    
    Attributes
    ----------
    _estimator_type : ``'classifier'``
        scikit-learn use this attribute to identify a classifier.

    _estimator_ensemble_type : ``'imbalanced_ensemble_classifier'``
        imbalanced-ensemble use this attribute to identify a classifier.
    """
    
    _estimator_type = "classifier"

    _estimator_ensemble_type = "imbalanced_ensemble_classifier"

    
    def _evaluate(self, 
                dataset_name:str, 
                eval_metrics:dict=None,
                return_header:bool=False,
                return_value_dict:bool=False,) -> str or dict:
        """Private function for performance evaluation during the 
        ensemble training process.
        """
        
        eval_datasets_ = self.eval_datasets_
        classes_ = self.classes_
        verbose_format_ = self.train_verbose_format_

        # Temporarily disable verbose
        support_verbose = hasattr(self, 'verbose')
        if support_verbose:
            verbose, self.verbose = self.verbose, 0

        # If no eval_metrics is given, use self.eval_metrics_
        if eval_metrics == None:
            eval_metrics = self.eval_metrics_

        # If return numerical results
        if return_value_dict == True:
            value_dict = {}
            for data_name, (X_eval, y_eval) in eval_datasets_.items():
                y_predict_proba = self.predict_proba(X_eval)
                data_value_dict = {}
                for metric_name, (metric_func, kwargs, ac_proba, ac_labels) \
                                                    in eval_metrics.items():
                    if ac_labels: kwargs['labels'] = classes_
                    if ac_proba: # If the metric take predict probabilities
                        score = metric_func(y_eval, y_predict_proba, **kwargs)
                    else: # If the metric do not take predict probabilities
                        y_predict = classes_.take(np.argmax(
                            y_predict_proba, axis=1), axis=0)
                        score = metric_func(y_eval, y_predict, **kwargs)
                    data_value_dict[metric_name] = score
                value_dict[data_name] = data_value_dict
            out = value_dict
        
        # If return string
        else:
            eval_info = ""
            if return_header == True:
                for metric_name in eval_metrics.keys():
                    eval_info = self._training_log_add_block(
                        eval_info, metric_name, "", "", " ", 
                        verbose_format_['len_metrics'][metric_name], strip=False)
            else:
                (X_eval, y_eval) = eval_datasets_[dataset_name]
                y_predict_proba = self.predict_proba(X_eval)
                for metric_name, (metric_func, kwargs, ac_proba, ac_labels) \
                                                    in eval_metrics.items():
                    if ac_labels: kwargs['labels'] = classes_
                    if ac_proba: # If the metric take predict probabilities
                        score = metric_func(y_eval, y_predict_proba, **kwargs)
                    else: # If the metric do not take predict probabilities
                        y_predict = classes_.take(np.argmax(
                            y_predict_proba, axis=1), axis=0)
                        score = metric_func(y_eval, y_predict, **kwargs)
                    eval_info = self._training_log_add_block(
                        eval_info, "{:.3f}".format(score), "", "", " ", 
                        verbose_format_['len_metrics'][metric_name], strip=False)
            out = eval_info[:-1]

        # Recover verbose state
        if support_verbose:
            self.verbose = verbose

        return out


    def _init_training_log_format(self):
        """Private function for initialization of the training verbose format"""

        if self.train_verbose_:
            len_iter = max(
                len(str(self.n_estimators)),
                len(TRAINING_LOG_HEAD_TITLES['iter'])) + 2
            if self.train_verbose_['print_distribution']:
                len_class_distr = max(
                    len(str(self.target_distr_)),
                    len(str(self.origin_distr_)),
                    len(TRAINING_LOG_HEAD_TITLES['class_distr'])) + 2
            else: len_class_distr = 0
            len_metrics = {
                metric_name: max(len(metric_name), 5) + 2
                for metric_name in self.eval_metrics_.keys()
            }
            metrics_total_length = sum(len_metrics.values()) + len(len_metrics) - 1
            len_datasets = {
                dataset_name: max(metrics_total_length, len("Data: "+dataset_name)+2)
                for dataset_name in self.eval_datasets_.keys()
            }
            self.train_verbose_format_ = {
                'len_iter': len_iter,
                'len_class_distr': len_class_distr,
                'len_metrics': len_metrics,
                'len_datasets': len_datasets,}

        return


    def _training_log_add_block(self, info, text, sta_char, fill_char, 
                                end_char, width, strip=True):
        """Private function for adding a block to training log."""

        info = info.rstrip(end_char) if strip else info
        info += "{}{:{fill}^{width}s}{}".format(
            sta_char, text, end_char,
            fill=fill_char, width=width)

        return info


    def _training_log_add_line(self, info="", texts=None, tabs=None, 
                               widths=None, flags=None):
        """Private function for adding a line to training log."""

        if texts == None:
            texts = ("", "", tuple("" for _ in self.eval_datasets_.keys()))
        if tabs == None:
            tabs = ("┃", "┃", "┃", " ")
        if widths == None:
            widths = (
                self.train_verbose_format_['len_iter'],
                self.train_verbose_format_['len_class_distr'],
                tuple(self.train_verbose_format_['len_datasets'].values())
            )
        if flags == None:
            flags = (True, self.train_verbose_['print_distribution'], self.train_verbose_['print_metrics'])
        (sta_char, mid_char, end_char, fill_char) = tabs
        (flag_iter, flag_distr, flag_metric) = flags
        (text_iter, text_distr, text_metrics) = texts
        (width_iter, width_distr, width_metrics) = widths
        if flag_iter:
            info = self._training_log_add_block(
                info, text_iter, sta_char, fill_char, end_char, width_iter)
        if flag_distr:
            info = self._training_log_add_block(
                info, text_distr, mid_char, fill_char, end_char, width_distr)
        if flag_metric:
            for text_metric, width_metric in zip(text_metrics, width_metrics):
                info = self._training_log_add_block(
                    info, text_metric, mid_char, fill_char, end_char, width_metric)

        return info
        

    def _training_log_to_console_head(self):
        """Private function for printing a table header."""

        # line 1
        info = self._training_log_add_line(
            tabs=("┏", "┳", "┓", "━"),
            )+"\n"
        # line 2
        info = self._training_log_add_line(info,
            texts=("", "", tuple("Data: "+data_name 
                for data_name in self.eval_datasets_.keys()))
            )+"\n"
        # line 3
        info = self._training_log_add_line(info,
            texts=(
                TRAINING_LOG_HEAD_TITLES['iter'], 
                TRAINING_LOG_HEAD_TITLES['class_distr'], 
                tuple("Metric" for data_name in self.eval_datasets_.keys())
                )
            )+"\n"
        # line 4
        info = self._training_log_add_line(info,
            texts=("", "", tuple(
                self._evaluate('', return_header=True)
                for data_name in self.eval_datasets_.keys()))
            )+"\n"
        # line 5
        info = self._training_log_add_line(info,
            tabs=("┣", "╋", "┫", "━"))

        return info


    def _training_log_to_console(self, i_iter=None, y=None):
        """Private function for printing training log to sys.stdout."""

        if self.train_verbose_:
            
            if not hasattr(self, '_properties'):
                raise AttributeError(
                    f"All imbalanced-ensemble estimators should" + \
                    f" have a `_properties` attribute to specify" + \
                    f" the method family they belong to."
                )

            try:
                training_type = self._properties['training_type']
            except Exception as e:
                e_args = list(e.args)
                e_args[0] += \
                    f" The key 'training_type' does not exist in" + \
                    f" the `_properties` attribute, please check" + \
                    f" your usage."
                e.args = tuple(e_args)
                raise e

            if training_type not in TRAINING_TYPES:
                raise ValueError(f"'training_type' should be in {TRAINING_TYPES}")
            if training_type == 'iterative':
                self._training_log_to_console_iterative(i_iter, y)
            elif training_type == 'parallel':
                self._training_log_to_console_parallel()
            else: raise NotImplementedError(
                f"'_training_log_to_console' for 'training_type' = {training_type}"
                f" needs to be implemented."
            )


    def _training_log_to_console_iterative(self, i_iter, y_resampled):
        """Private function for printing training log to sys.stdout.
        (for ensemble classifiers that train in an iterative manner)"""

        if i_iter == 0:
            print(self._training_log_to_console_head())
        
        eval_data_names = self.eval_datasets_.keys()

        if (i_iter+1) % self.train_verbose_['granularity'] == 0 or i_iter == 0:
            print(self._training_log_add_line(texts=(
                    f"{i_iter+1}", f"{sort_dict_by_key(Counter(y_resampled))}", 
                    tuple(self._evaluate(data_name) for data_name in eval_data_names)
                )))
        
        if (i_iter+1) == self.n_estimators:
            print(self._training_log_add_line(tabs=("┣", "╋", "┫", "━")))
            print(self._training_log_add_line(texts=(
                    "final", f"{dict(Counter(y_resampled))}", 
                    tuple(self._evaluate(data_name) for data_name in eval_data_names)
                )))
            print(self._training_log_add_line(tabs=("┗", "┻", "┛", "━")))


    def _training_log_to_console_parallel(self):
        """Private function for printing training log to sys.stdout.
        (for ensemble classifiers that train in a parallel manner)"""

        eval_data_names = self.eval_datasets_.keys()
        print(self._training_log_to_console_head())
        print(self._training_log_add_line(texts=(
                str(self.n_estimators), "", 
                tuple(self._evaluate(data_name) for data_name in eval_data_names)
            )))
        print(self._training_log_add_line(tabs=("┗", "┻", "┛", "━")))


_properties = {
    'ensemble_type': 'general',
}


class BaseImbalancedEnsemble(ImbalancedEnsembleClassifierMixin, 
                             BaseEnsemble, metaclass=ABCMeta):
    """Base class for DupleBalanceClassifier.

    Warning: This class should not be used directly. Use derived classes
    instead.

    Parameters
    ----------
    base_estimator : object
        The base estimator from which the ensemble is built.

    n_estimators : int, default=10
        The number of estimators in the ensemble.

    estimator_params : list of str, default=tuple()
        The list of attributes to use as parameters when instantiating a
        new base estimator. If none are given, default parameters are used.

    n_jobs : int, default=None
        The number of jobs to run in parallel for :meth:`predict`. 
        ``None`` means 1 unless in a :obj:`joblib.parallel_backend` 
        context. ``-1`` means using all processors. See `Glossary <https://scikit-learn.org/stable/glossary.html#term-n_jobs>`_
        for more details.
    
    random_state : int, RandomState instance or None, default=None
        Control the randomization of the algorithm.
        If the base estimator accepts a `random_state` attribute, a different
        seed is generated for each instance in the ensemble.
        Pass an ``int`` for reproducible output across multiple function calls.
        
        - If ``int``, ``random_state`` is the seed used by the random number
          generator;
        - If ``RandomState`` instance, random_state is the random number
          generator;
        - If ``None``, the random number generator is the ``RandomState``
          instance used by ``np.random``.

    verbose : int, default=0
        Controls the verbosity when predicting.

    Attributes
    ----------
    base_estimator_ : estimator
        The base estimator from which the ensemble is grown.
        
    estimators_ : list of estimators
        The collection of fitted base estimators.
    """
    
    def __init__(self, 
                 base_estimator, 
                 n_estimators=10,
                 estimator_params=tuple(), 
                 random_state=None,
                 n_jobs=None, 
                 verbose=0,):

        self.random_state = random_state
        self.n_jobs = n_jobs
        self.verbose = verbose
        self.check_x_y_args = {
            'accept_sparse': ['csr', 'csc'],
            'force_all_finite': False,
            'dtype': None,
        }

        super(BaseImbalancedEnsemble, self).__init__(
            base_estimator=base_estimator,
            n_estimators=n_estimators,
            estimator_params=estimator_params,
        )

        self._properties = _properties
    

    def _validate_y(self, y):
        """Validate the label vector."""
        y = column_or_1d(y, warn=True)
        check_classification_targets(y)
        self.classes_, y = np.unique(y, return_inverse=True)
        self.n_classes_ = len(self.classes_)
        return y


    def _validate_estimator(self, default):
        """Check the estimator, sampler and the n_estimator attribute.

        Sets the base_estimator_` and base_sampler_` attributes.
        """

        # validate estimator using 
        # sklearn.ensemble.BaseEnsemble._validate_estimator
        super()._validate_estimator(default=default)

        if hasattr(self, 'base_sampler'):
            # validate sampler and sampler_kwargs
            # validated sampler stored in self.base_sampler_
            try:
                self.base_sampler_ = clone(self.base_sampler)
            except Exception as e:
                e_args = list(e.args)
                e_args[0] = "Exception occurs when trying to validate" + \
                            " base_sampler: " + e_args[0]
                e.args = tuple(e_args)
                raise e
    

    def _make_sampler(self, append=True, random_state=None, **overwrite_kwargs):
        """Make and configure a copy of the `base_sampler_` attribute.

        Warning: This method should be used to properly instantiate new
        sub-samplers.
        """

        sampler = clone(self.base_sampler_)
        if hasattr(self, 'sampler_kwargs_'):
            sampler.set_params(**self.sampler_kwargs_)

        # Arguments passed to _make_sampler function have higher priority,
        # they will overwrite the self.sampler_kwargs_
        sampler.set_params(**overwrite_kwargs)
        
        if random_state is not None:
            _set_random_states(sampler, random_state)

        if append:
            self.samplers_.append(sampler)

        return sampler
    

    @_deprecate_positional_args
    def fit(self, X, y, *, sample_weight=None, **kwargs):
        """Build the ensemble classifier from the training set (X, y)."""

        # Check random state
        self.random_state = check_random_state(self.random_state)

        # Convert data (X is required to be 2d and indexable)
        X, y = self._validate_data(X, y, **self.check_x_y_args)
        if sample_weight is not None:
            sample_weight = _check_sample_weight(sample_weight, X, dtype=np.float64)
            sample_weight /= sample_weight.sum()
            if np.any(sample_weight < 0):
                raise ValueError("sample_weight cannot contain negative weights")

        # Remap output
        n_samples, self.n_features_ = X.shape
        self.features_ = np.arange(self.n_features_)
        self._n_samples = n_samples
        y = self._validate_y(y)

        # Check parameters
        self._validate_estimator(default=DecisionTreeClassifier())
        
        # If the base estimator do not support sample weight and sample weight
        # is not None, raise an ValueError
        support_sample_weight = has_fit_parameter(self.base_estimator_,
                                                "sample_weight")
        if not support_sample_weight and sample_weight is not None:
            raise ValueError("The base estimator doesn't support sample weight")

        self.estimators_, self.estimators_features_ = [], []

        return self._fit(X, y, sample_weight=sample_weight, **kwargs)
    

    @abstractmethod
    def _fit(self, X, y, sample_weight, **kwargs):
        """Needs to be implemented in the derived class"""
        return self
        
    
    def predict_proba(self, X):
        """Predict class probabilities for X.

        The predicted class probabilities of an input sample is computed as
        the mean predicted class probabilities of the base estimators in the
        ensemble. If base estimators do not implement a ``predict_proba``
        method, then it resorts to voting and the predicted class probabilities
        of an input sample represents the proportion of estimators predicting
        each class.

        Parameters
        ----------
        X : {array-like, sparse matrix} of shape = [n_samples, n_features]
            The training input samples. Sparse matrices are accepted only if
            they are supported by the base estimator.

        Returns
        -------
        p : array of shape = [n_samples, n_classes]
            The class probabilities of the input samples. 
        """

        check_is_fitted(self)
        # Check data
        X = check_array(
            X, accept_sparse=['csr', 'csc'], dtype=None,
            force_all_finite=False
        )
        if self.n_features_ != X.shape[1]:
            raise ValueError("Number of features of the model must "
                             "match the input. Model n_features is {0} and "
                             "input n_features is {1}."
                             "".format(self.n_features_, X.shape[1]))
        
        # Parallel loop
        n_jobs, _, starts = _partition_estimators(self.n_estimators,
                                                             self.n_jobs)

        all_proba = Parallel(n_jobs=n_jobs, verbose=self.verbose,
                             **self._parallel_args())(
            delayed(_parallel_predict_proba)(
                self.estimators_[starts[i]:starts[i + 1]],
                self.estimators_features_[starts[i]:starts[i + 1]],
                X,
                self.n_classes_)
            for i in range(n_jobs))

        # Reduce
        proba = sum(all_proba) / len(self.estimators_)

        return proba
        

    def predict(self, X):
        """Predict class for X.

        The predicted class of an input sample is computed as the class with
        the highest mean predicted probability. If base estimators do not
        implement a ``predict_proba`` method, then it resorts to voting.

        Parameters
        ----------
        X : {array-like, sparse matrix} of shape (n_samples, n_features)
            The training input samples. Sparse matrices are accepted only if
            they are supported by the base estimator.

        Returns
        -------
        y : ndarray of shape (n_samples,)
            The predicted classes.
        """
        
        predicted_probabilitiy = self.predict_proba(X)
        return self.classes_.take((np.argmax(predicted_probabilitiy, axis=1)),
                                  axis=0)


    @property
    def feature_importances_(self):
        """The impurity-based feature importances.
        The higher, the more important the feature.
        The importance of a feature is computed as the (normalized)
        total reduction of the criterion brought by that feature.  It is also
        known as the Gini importance.
        Warning: impurity-based feature importances can be misleading for
        high cardinality features (many unique values). See
        :func:`sklearn.inspection.permutation_importance` as an alternative.
        Returns
        -------
        feature_importances_ : ndarray of shape (n_features,)
            The feature importances.
        """
        if self.estimators_ is None or len(self.estimators_) == 0:
            raise ValueError(
                "Estimator not fitted, call `fit` before `feature_importances_`."
            )

        try:
            if hasattr(self, 'estimator_weights_'):
                norm = self.estimator_weights_.sum()
                return (
                    sum(
                        weight * clf.feature_importances_
                        for weight, clf in zip(self.estimator_weights_, self.estimators_)
                    )
                    / norm
                )
            else:
                return (
                    sum(
                        clf.feature_importances_ for clf in self.estimators_
                    )
                    / len(self.estimators_)
                )

        except AttributeError as e:
            raise AttributeError(
                "Unable to compute feature importances "
                "since base_estimator does not have a "
                "feature_importances_ attribute"
            ) from e


    def _parallel_args(self):
        return {}

# %%



"""
``self-paced ensemble`` is a python-based ensemble learning framework for 
dealing with binary class-imbalanced classification problems in machine learning.

Subpackages
-----------
self_paced_ensemble
    Module which provides our SelfPacedEnsembleClassifier implementation.
canonical_ensemble
    Module which provides baseline methods based on ensemble learning.
canonical_resampling
    Module which provides baseline methods based on data resampling.
utils
    Module including various utilities.
"""

from . import self_paced_ensemble
from . import canonical_ensemble
from . import canonical_resampling
from . import utils

from .self_paced_ensemble import SelfPacedEnsembleClassifier

from .__version__ import __version__

__all__ = [
    "SelfPacedEnsembleClassifier",
    "self_paced_ensemble",
    "canonical_ensemble",
    "canonical_resampling",
    "utils",
    "__version__",
]


"""Utilities for data validation."""

# Authors: Zhining Liu <zhining.liu@outlook.com>
# License: MIT


from collections import OrderedDict
from sklearn.utils import check_X_y


VALID_DATA_INFO = "'eval_datasets' should be a `dict` of validation data," + \
                  " e.g., {..., dataset_name : (X_valid, y_valid), ...}."

TRAIN_DATA_NAME = "train"


def _check_eval_datasets_name(data_name):
    if not isinstance(data_name, str):
        raise TypeError(
            VALID_DATA_INFO + \
            f" The keys must be `string`, got {type(data_name)}, " + \
            f" please check your usage."
        )
    if data_name == TRAIN_DATA_NAME:
        raise ValueError(
            f"The name {TRAIN_DATA_NAME} is reserved for the training"
            f" data (it will automatically add into the 'eval_datasets_'"
            f" attribute after calling `fit`), please use another name"
            f" for your evaluation dataset."
        )
    return data_name


def _check_eval_datasets_tuple(data_tuple, data_name, **check_x_y_kwargs):
    if not isinstance(data_tuple, tuple):
        raise TypeError(
            VALID_DATA_INFO + \
            f" The value of '{data_name}' is {type(data_tuple)} (should be tuple)," + \
            f" please check your usage."
        )
    elif len(data_tuple) != 2:
        raise ValueError(
            VALID_DATA_INFO + \
            f" The data tuple of '{data_name}' has {len(data_tuple)} element(s)" + \
            f" (should be 2), please check your usage."
        )
    else:
        X, y =  check_X_y(data_tuple[0], data_tuple[1], **check_x_y_kwargs)
        return (X, y)


def _check_eval_datasets_dict(eval_datasets_dict, **check_x_y_kwargs):

    if TRAIN_DATA_NAME in eval_datasets_dict.keys():
        raise ValueError(
            f"The name '{TRAIN_DATA_NAME}' could not be used"
            f" for the validation datasets. Please use another name."
        )

    eval_datasets_dict_ = {}
    for data_name, data_tuple in eval_datasets_dict.items():
        data_name_ = _check_eval_datasets_name(data_name)
        data_tuple_ = _check_eval_datasets_tuple(data_tuple, data_name_, **check_x_y_kwargs)
        eval_datasets_dict_[data_name_] = data_tuple_
    
    return eval_datasets_dict_


def _all_elements_equal(list_to_check:list) -> bool:
    if len(list_to_check) == 1:
        return True
    return all([
        (list_to_check[i] == list_to_check[i+1]) 
        for i in range(len(list_to_check)-1)
        ])


def check_eval_datasets(eval_datasets, X_train_=None, y_train_=None, **check_x_y_kwargs):
    """Check `eval_datasets` parameter."""
    # Whether to add training data in to returned data dictionary
    if X_train_ is None and y_train_ is None:
        result_datasets = OrderedDict({})
    else:
        result_datasets = OrderedDict({TRAIN_DATA_NAME: (X_train_, y_train_)})

    # If eval_datasets is None
    #   return data dictionary
    if eval_datasets == None:
        return result_datasets

    # If eval_datasets is dict
    elif isinstance(eval_datasets, dict):

        # Check dict and validate all names (keys) and data tuples (values)
        eval_datasets_ = _check_eval_datasets_dict(eval_datasets, **check_x_y_kwargs)

        # Combine train_datasets and eval_datasets_
        result_datasets.update(eval_datasets_)

        # Ensure all datasets have the same number of features
        if not _all_elements_equal([data_tuple[0].shape[1]
                                    for data_tuple in result_datasets.values()]):
            raise ValueError(
                f"The train + evaluation datasets have inconsistent number of"
                f" features. Make sure that the data given in 'eval_datasets'"
                f" and the training data ('X', 'y') are sampled from the same"
                f" task/distribution."
            )
        return result_datasets

    # Else raise TypeError
    else:
        raise TypeError(
            VALID_DATA_INFO + \
            f" Got {type(eval_datasets)}, please check your usage."
        )


# -*- coding: utf-8 -*-
"""
A resampling-based classifier for imbalanced classification.
15 resampling algorithms are included: 
'RUS', 'CNN', 'ENN', 'NCR', 'Tomek', 'ALLKNN', 'OSS',
'NM', 'CC', 'SMOTE', 'ADASYN', 'BorderSMOTE', 'SMOTEENN', 
'SMOTETomek', 'ORG'.

The implementation of these resampling algorithms is based on `imblearn`.
Please refer to https://github.com/scikit-learn-contrib/imbalanced-learn.
"""

# Created on Sun Jan 13 14:32:27 2019
# Authors: Zhining Liu <zhining.liu@outlook.com>
# License: MIT

from imblearn.under_sampling import (
    ClusterCentroids, 
    NearMiss, 
    RandomUnderSampler, 
    EditedNearestNeighbours, 
    AllKNN, 
    TomekLinks, 
    OneSidedSelection, 
    CondensedNearestNeighbour, 
    NeighbourhoodCleaningRule,
)
from imblearn.over_sampling import (
    SMOTE, ADASYN, BorderlineSMOTE,
)
from imblearn.combine import (
    SMOTEENN, SMOTETomek,
)

from sklearn.tree import DecisionTreeClassifier as DT


SUPPORT_RESAMPLING = ['RUS', 'CNN', 'ENN', 'NCR', 'Tomek', 'ALLKNN', 'OSS',
                'NM', 'CC', 'SMOTE', 'ADASYN', 'BorderSMOTE', 'SMOTEENN', 
                'SMOTETomek', 'ORG']


class Error(Exception):
    pass

class ResampleClassifier(object):
    '''
    Re-sampling methods for imbalance classification, based on imblearn python package.
    imblearn url: https://github.com/scikit-learn-contrib/imbalanced-learn
    Hyper-parameters:
        base_estimator : scikit-learn classifier object
            optional (default=DecisionTreeClassifier)
            The base estimator used for training after re-sampling
    '''
    def __init__(self, base_estimator=DT()):
        self.base_estimator = base_estimator
    
    def predict(self, X):
        return self.base_estimator.predict(X)
    
    def fit(self, X, y, by, random_state=None, visualize=False):
        '''
        by: String
            The method used to perform re-sampling
            support: ['RUS', 'CNN', 'ENN', 'NCR', 'Tomek', 'ALLKNN', 'OSS',
                'NM', 'CC', 'SMOTE', 'ADASYN', 'BorderSMOTE', 'SMOTEENN', 
                'SMOTETomek', 'ORG']
        '''
        if by == 'RUS':
            sampler = RandomUnderSampler(random_state=random_state)
        elif by == 'CNN':
            sampler = CondensedNearestNeighbour(random_state=random_state)
        elif by == 'ENN':
            sampler = EditedNearestNeighbours()
        elif by == 'NCR':
            sampler = NeighbourhoodCleaningRule()
        elif by == 'Tomek':
            sampler = TomekLinks()
        elif by == 'ALLKNN':
            sampler = AllKNN()
        elif by == 'OSS':
            sampler = OneSidedSelection(random_state=random_state)
        elif by == 'NM':
            sampler = NearMiss()
        elif by == 'CC':
            sampler = ClusterCentroids(random_state=random_state)
        elif by == 'SMOTE':
            sampler = SMOTE(random_state=random_state)
        elif by == 'ADASYN':
            sampler = ADASYN(random_state=random_state)
        elif by == 'BorderSMOTE':
            sampler = BorderlineSMOTE(random_state=random_state)
        elif by == 'SMOTEENN':
            sampler = SMOTEENN(random_state=random_state)
        elif by == 'SMOTETomek':
            sampler = SMOTETomek(random_state=random_state)
        elif by == 'ORG':
            sampler = None
        else:
            raise Error('Unexpected \'by\' type {}'.format(by))
        
        if by != 'ORG':
            X_train, y_train = sampler.fit_resample(X, y)
        else:
            X_train, y_train = X, y
        self.base_estimator.fit(X_train, y_train)


# -*- coding: utf-8 -*-
"""
Five ensemble learning algorithms for imbalanced classification, including:
'SMOTEBaggingClassifier', 'SMOTEBoostClassifier', 'RUSBoostClassifier', 
'UnderBaggingClassifier', and 'BalanceCascadeClassifier'.

Note: methods in this module are now included in the `imbalanced-ensemble`.
Please refer to https://imbalanced-ensemble.readthedocs.io/ for more details.
"""

# Created on Sun Jan 13 14:32:27 2019
# Authors: Zhining Liu <zhining.liu@outlook.com>
# License: MIT

# %%

from imbens.ensemble import SMOTEBoostClassifier
from imbens.ensemble import SMOTEBaggingClassifier
from imbens.ensemble import RUSBoostClassifier
from imbens.ensemble import UnderBaggingClassifier
from imbens.ensemble import BalanceCascadeClassifier


"""
--------------------------------------------------------------------------
The `self-paced-ensemble.utils` module implement various utilities.
--------------------------------------------------------------------------
"""

from ._utils import (
    load_covtype_dataset,
    make_binary_classification_target,
    imbalance_train_test_split,
    imbalance_random_subset,
    auc_prc,
    f1_optim,
    gm_optim,
    mcc_optim,
    save_model,
    load_model,
)

__all__ = [
    "load_covtype_dataset",
    "make_binary_classification_target",
    "imbalance_train_test_split",
    "imbalance_random_subset",
    "auc_prc",
    "f1_optim",
    "gm_optim",
    "mcc_optim",
    "save_model",
    "load_model",
]



"""
--------------------------------------------------------------------------
The `self_paced_ensemble.canonical_resampling` module implement a 
resampling-based classifier for imbalanced classification.
15 resampling algorithms are included: 
'RUS', 'CNN', 'ENN', 'NCR', 'Tomek', 'ALLKNN', 'OSS',
'NM', 'CC', 'SMOTE', 'ADASYN', 'BorderSMOTE', 'SMOTEENN', 
'SMOTETomek', 'ORG'.

Note: the implementation of these resampling algorithms is based on 
imblearn python package. 
See https://github.com/scikit-learn-contrib/imbalanced-learn.
--------------------------------------------------------------------------
"""


from .canonical_resampling import ResampleClassifier

__all__ = [
    "ResampleClassifier",
]



"""Class to perform self-paced under-sampling."""

# Authors: Zhining Liu <zhining.liu@outlook.com>
# License: MIT

# %%

import numbers
import numpy as np

from sklearn.utils import check_random_state
from sklearn.utils import _safe_indexing

from ._base_sampler import BaseSampler
from ..utils._validation_param import check_pred_proba, check_type
from ..utils._validation import _deprecate_positional_args, check_target_type

# # For local test
# import sys
# sys.path.append("..")
# from self_paced_ensemble._base_sampler import BaseSampler
# from utils._validation_param import check_pred_proba, check_type
# from utils._validation import _deprecate_positional_args, check_target_type

# %%

class SelfPacedUnderSampler(BaseSampler):
    """Class to perform self-paced under-sampling in [1]_.

    Parameters
    ----------
    sampling_strategy : float, str, dict, callable, default='auto'
        Sampling information to sample the data set.

        - When ``float``, it corresponds to the desired ratio of the number of
          samples in the minority class over the number of samples in the
          majority class after resampling. Therefore, the ratio is expressed as
          :math:`\alpha_{us} = N_{m} / N_{rM}` where :math:`N_{m}` is the
          number of samples in the minority class and
          :math:`N_{rM}` is the number of samples in the majority class
          after resampling.

          .. warning::
             ``float`` is only available for **binary** classification. An
             error is raised for multi-class classification.

        - When ``str``, specify the class targeted by the resampling. The
          number of samples in the different classes will be equalized.
          Possible choices are:

            ``'majority'``: resample only the majority class;

            ``'not minority'``: resample all classes but the minority class;

            ``'not majority'``: resample all classes but the majority class;

            ``'all'``: resample all classes;

            ``'auto'``: equivalent to ``'not minority'``.

        - When ``dict``, the keys correspond to the targeted classes. The
          values correspond to the desired number of samples for each targeted
          class.

        - When callable, function taking ``y`` and returns a ``dict``. The keys
          correspond to the targeted classes. The values correspond to the
          desired number of samples for each class.

    k_bins : int, default=5
        The number of hardness bins that were used to approximate 
        hardness distribution. It is recommended to set it to 5. 
        One can try a larger value when the smallest class in the 
        data set has a sufficient number (say, > 1000) of samples.

    soft_resample_flag : bool, default=False
        Whether to use weighted sampling to perform soft self-paced 
        under-sampling, rather than explicitly cut samples into 
        ``k``-bins and perform hard sampling.

    replacement : bool, default=True
        Whether samples are drawn with replacement. If ``False`` 
        and ``soft_resample_flag = False``, may raise an error when 
        a bin has insufficient number of data samples for resampling.

    random_state : int, RandomState instance, default=None
        Control the randomization of the algorithm.

        - If ``int``, ``random_state`` is the seed used by the random number
          generator;
        - If ``RandomState`` instance, random_state is the random number
          generator;
        - If ``None``, the random number generator is the ``RandomState``
          instance used by ``np.random``.

    Attributes
    ----------
    sample_indices_ : ndarray of shape (n_new_samples,)
        Indices of the samples selected.

    See Also
    --------
    BalanceCascadeUnderSampler :  Dynamic under-sampling for BalanceCascade.

    Notes
    -----
    Supports multi-class resampling by sampling each class independently.
    Supports heterogeneous data as object array containing string and numeric
    data.
    
    References
    ----------
    .. [1] Liu, Z., Cao, W., Gao, Z., Bian, J., Chen, H., Chang, Y., & Liu, T. Y. 
       "Self-paced ensemble for highly imbalanced massive data classification." 
       2020 IEEE 36th International Conference on Data Engineering (ICDE). 
       IEEE, 2010: 841-852.
    """

    _sampling_type = 'under-sampling'

    @_deprecate_positional_args
    def __init__(
        self, *, 
        sampling_strategy="auto", 
        k_bins=5,
        soft_resample_flag=True, 
        replacement=False, 
        random_state=None, 
    ):
        super().__init__(sampling_strategy=sampling_strategy)

        self.k_bins = k_bins
        self.soft_resample_flag = soft_resample_flag
        self.replacement = replacement
        self.random_state = random_state

        # Check parameters
        self.k_bins_ = check_type(k_bins, 'k_bins', numbers.Integral)
        self.replacement_ = check_type(replacement, 'replacement', bool)
        self.soft_resample_flag_ = check_type(soft_resample_flag, 
            'soft_resample_flag', bool)


    def _check_X_y(self, X, y):
        y, binarize_y = check_target_type(y, indicate_one_vs_all=True)
        X, y = self._validate_data(
            X,
            y,
            reset=True,
            accept_sparse=["csr", "csc"],
            dtype=None,
            force_all_finite=False,
        )
        return X, y, binarize_y


    def fit_resample(self, X, y, *, sample_weight, **kwargs):
        """Resample the dataset.

        Parameters
        ----------
        X : {array-like, dataframe, sparse matrix} of shape \
                (n_samples, n_features)
            Matrix containing the data which have to be sampled.

        y : array-like of shape (n_samples,)
            Corresponding label for each sample in X.

        y_pred_proba : array-like of shape (n_samples, n_classes)
            The predicted class probabilities of the input samples
            by the current SPE ensemble classifier. The order of the
            classes corresponds to that in the parameter `classes_`.
        
        alpha : float
            The self-paced factor that controls SPE under-sampling.
        
        classes_ : ndarray of shape (n_classes,)
            The classes labels.

        sample_weight : array-like of shape (n_samples,), default=None
            Corresponding weight for each sample in X.
        
        Returns
        -------
        X_resampled : {array-like, dataframe, sparse matrix} of shape \
                (n_samples_new, n_features)
            The array containing the resampled data.

        y_resampled : array-like of shape (n_samples_new,)
            The corresponding label of `X_resampled`.
        
        sample_weight : array-like of shape (n_samples_new,), default=None
            The corresponding weight of `X_resampled`.
            Only will be returned if input sample_weight is not None.
        """
        return super().fit_resample(X, y, sample_weight=sample_weight, **kwargs)


    @_deprecate_positional_args
    def _fit_resample(self, X, y, *, 
                      y_pred_proba, alpha:float, 
                      classes_, sample_weight=None):
        
        n_samples, n_classes = X.shape[0], classes_.shape[0]

        # Check random_state and predict probabilities
        random_state = check_random_state(self.random_state)
        y_pred_proba = check_pred_proba(y_pred_proba, n_samples, n_classes, dtype=np.float64)

        # Check the self-paced factor alpha
        alpha = check_type(alpha, 'alpha', numbers.Number)
        if alpha < 0:
            raise ValueError("'alpha' must not be negative.")

        indexes = np.arange(n_samples)
        index_list = []

        # For each class C
        for target_class in classes_:
            if target_class in self.sampling_strategy_.keys():

                # Get the desired & actual number of samples of class C
                # and the index mask of class C
                n_target_samples_c = self.sampling_strategy_[target_class]
                class_index_mask = y == target_class
                n_samples_c = np.count_nonzero(class_index_mask)

                # Compute the hardness array
                hardness_c=np.abs(
                    np.ones(n_samples_c) - \
                    y_pred_proba[class_index_mask, target_class])

                # index_c: absolute indexes of class C samples
                index_c = indexes[class_index_mask]

                if n_target_samples_c <= n_samples_c:
                    
                    # Get the absolute indexes of resampled class C samples
                    index_c_result = self._undersample_single_class(
                        hardness_c=hardness_c,
                        n_target_samples_c=n_target_samples_c,
                        index_c=index_c,
                        alpha=alpha,
                        random_state=random_state)

                # If no sufficient samples in class C, raise an RuntimeError
                else: raise RuntimeError(
                    f"Got n_target_samples_c ({n_target_samples_c})"
                    f" > n_samples_c ({n_samples_c} for class {target_class}.)"
                )
                index_list.append(index_c_result)
        
        # Concatenate the result
        index_spu = np.hstack(index_list)

        # Store the final undersample indexes
        self.sample_indices_ = index_spu

        # Return the resampled X, y
        # also return resampled sample_weight (if sample_weight is not None)
        if sample_weight is not None:
            # sample_weight is already validated in super().fit_resample()
            weights_under = _safe_indexing(sample_weight, index_spu)
            return _safe_indexing(X, index_spu), _safe_indexing(y, index_spu), weights_under
        else: return _safe_indexing(X, index_spu), _safe_indexing(y, index_spu)


    def _undersample_single_class(self, hardness_c, n_target_samples_c, 
                                  index_c, alpha, random_state):
        """Perform self-paced under-sampling in a single class"""
        k_bins = self.k_bins_
        soft_resample_flag = self.soft_resample_flag_
        replacement = self.replacement_
        n_samples_c = hardness_c.shape[0]

        # if hardness is not distinguishable or no sample will be dropped
        if hardness_c.max() == hardness_c.min() or n_target_samples_c == n_samples_c:
            # perform random under-sampling
            return random_state.choice(
                index_c,
                size=n_target_samples_c,
                replace=replacement)

        with np.errstate(divide='ignore', invalid='ignore'):
            # compute population & hardness contribution of each bin
            populations, edges = np.histogram(hardness_c, bins=k_bins)
            contributions = np.zeros(k_bins)
            index_bins = []
            for i_bin in range(k_bins):
                index_bin = ((hardness_c >= edges[i_bin]) & (hardness_c < edges[i_bin+1]))
                if i_bin == (k_bins-1):
                    index_bin = index_bin | (hardness_c==edges[i_bin+1])
                index_bins.append(index_bin)
                if populations[i_bin] > 0:
                    contributions[i_bin] = hardness_c[index_bin].mean()

            # compute the expected number of samples to be sampled from each bin
            bin_weights = 1. / (contributions + alpha)
            bin_weights[np.isnan(bin_weights)|np.isinf(bin_weights)] = 0
            n_target_samples_bins = n_target_samples_c * bin_weights / bin_weights.sum()
            # check whether exists empty bins
            n_invalid_samples = sum(n_target_samples_bins[populations==0])
            if n_invalid_samples > 0:
                n_valid_samples = n_target_samples_c-n_invalid_samples
                n_target_samples_bins *= n_target_samples_c / n_valid_samples
                n_target_samples_bins[populations==0] = 0
            n_target_samples_bins = n_target_samples_bins.astype(int)+1

        if soft_resample_flag:
            with np.errstate(divide='ignore', invalid='ignore'):
                # perform soft (weighted) self-paced under-sampling
                soft_spu_bin_weights = n_target_samples_bins / populations
                soft_spu_bin_weights[~np.isfinite(soft_spu_bin_weights)] = 0
            # compute sampling probabilities
            soft_spu_sample_proba = np.zeros_like(hardness_c)
            for i_bin in range(k_bins):
                soft_spu_sample_proba[index_bins[i_bin]] = soft_spu_bin_weights[i_bin]
            soft_spu_sample_proba /= soft_spu_sample_proba.sum()
            # sample with respect to the sampling probabilities
            return random_state.choice(
                index_c,
                size=n_target_samples_c,
                replace=replacement,
                p=soft_spu_sample_proba,)
        else:
            # perform hard self-paced under-sampling
            index_c_results = []
            for i_bin in range(k_bins):
                # if no sufficient data in bin for resampling, raise an RuntimeError
                if populations[i_bin] < n_target_samples_bins[i_bin] and not replacement:
                    raise RuntimeError(
                        f"Met {i_bin}-th bin with insufficient number of data samples"
                        f" ({populations[i_bin]}, expected"
                        f" >= {n_target_samples_bins[i_bin]})."
                        f" Set 'soft_resample_flag' or 'replacement' to `True` to."
                        f" avoid this issue."
                    )
                index_c_bin = index_c[index_bins[i_bin]]
                # random sample from each bin
                if len(index_c_bin) > 0:
                    index_c_results.append(
                        random_state.choice(
                            index_c_bin,
                            size=n_target_samples_bins[i_bin],
                            replace=replacement,)
                    )
            # concatenate and return the result
            index_c_result = np.hstack(index_c_results)
            return index_c_result
        

    def _more_tags(self):
        return {
            "X_types": ["2darray", "string", "sparse", "dataframe"],
            "sample_indices": True,
            "allow_nan": True,
        }

# %%

if __name__ == "__main__":
    
    from collections import Counter
    import numpy as np

    from sklearn.utils import check_random_state
    from sklearn.utils import _safe_indexing
    from sklearn.datasets import make_classification
    from sklearn.ensemble import BaggingClassifier

    X, y = make_classification(n_classes=3, class_sep=2,
        weights=[0.1, 0.3, 0.6], n_informative=3, n_redundant=1, flip_y=0,
        n_features=20, n_clusters_per_class=1, n_samples=1000, random_state=10)
    print('Original dataset shape %s' % Counter(y))

    sampling_strategy_ = {2: 200, 1: 100, 0: 90}
    print('Target dataset shape %s' % sampling_strategy_)

    sample_weight = np.full_like(y, fill_value=1/y.shape[0], dtype=float)
    clf = BaggingClassifier(
        n_estimators=50,
    ).fit(X, y)
    
    y_pred_proba = clf.predict_proba(X)

    alpha = 0

    spu = SelfPacedUnderSampler(
        sampling_strategy=sampling_strategy_,
        k_bins=5,
        soft_resample_flag=True,
        replacement=False, 
        random_state=0,
    )
    X_res, y_res, weights_res = spu.fit_resample(X, y, 
        y_pred_proba=y_pred_proba, alpha=0, sample_weight=sample_weight,
        classes_=clf.classes_)
    print('Resampled dataset shape %s' % Counter(y_res))

# %%



"""Utilities for input validation."""

# Adapted from imbalanced-learn

# Authors: Guillaume Lemaitre <g.lemaitre58@gmail.com>
# License: MIT

import warnings
from collections import OrderedDict
from functools import wraps
from inspect import signature, Parameter
from numbers import Integral, Real

import numpy as np

from sklearn.utils import column_or_1d
from sklearn.utils.multiclass import type_of_target


SAMPLING_KIND = (
    "over-sampling",
    "under-sampling",
    "clean-sampling",
    "hybrid-sampling",
    "ensemble",
    "bypass",
)
TARGET_KIND = ("binary", "multiclass", "multilabel-indicator")


class ArraysTransformer:
    """A class to convert sampler output arrays to their original types."""

    def __init__(self, X, y):
        self.x_props = self._gets_props(X)
        self.y_props = self._gets_props(y)

    def transform(self, X, y):
        X = self._transfrom_one(X, self.x_props)
        y = self._transfrom_one(y, self.y_props)
        return X, y

    def _gets_props(self, array):
        props = {}
        props["type"] = array.__class__.__name__
        props["columns"] = getattr(array, "columns", None)
        props["name"] = getattr(array, "name", None)
        props["dtypes"] = getattr(array, "dtypes", None)
        return props

    def _transfrom_one(self, array, props):
        type_ = props["type"].lower()
        if type_ == "list":
            ret = array.tolist()
        elif type_ == "dataframe":
            import pandas as pd

            ret = pd.DataFrame(array, columns=props["columns"])
            ret = ret.astype(props["dtypes"])
        elif type_ == "series":
            import pandas as pd

            ret = pd.Series(array, dtype=props["dtypes"], name=props["name"])
        else:
            ret = array
        return ret


def _count_class_sample(y):
    unique, counts = np.unique(y, return_counts=True)
    return dict(zip(unique, counts))


def check_target_type(y, indicate_one_vs_all=False):
    """Check the target types to be conform to the current samplers.

    The current samplers should be compatible with ``'binary'``,
    ``'multilabel-indicator'`` and ``'multiclass'`` targets only.

    Parameters
    ----------
    y : ndarray
        The array containing the target.

    indicate_one_vs_all : bool, default=False
        Either to indicate if the targets are encoded in a one-vs-all fashion.

    Returns
    -------
    y : ndarray
        The returned target.

    is_one_vs_all : bool, optional
        Indicate if the target was originally encoded in a one-vs-all fashion.
        Only returned if ``indicate_multilabel=True``.
    """
    type_y = type_of_target(y)
    if type_y == "multilabel-indicator":
        if np.any(y.sum(axis=1) > 1):
            raise ValueError(
                "Imbalanced-learn currently supports binary, multiclass and "
                "binarized encoded multiclasss targets. Multilabel and "
                "multioutput targets are not supported."
            )
        y = y.argmax(axis=1)
    else:
        y = column_or_1d(y)

    return (y, type_y == "multilabel-indicator") if indicate_one_vs_all else y


def _sampling_strategy_all(y, sampling_type):
    """Returns sampling target by targeting all classes."""
    target_stats = _count_class_sample(y)
    if sampling_type == "over-sampling":
        n_sample_majority = max(target_stats.values())
        sampling_strategy = {
            key: n_sample_majority - value for (key, value) in target_stats.items()
        }
    elif sampling_type == "under-sampling" or sampling_type == "clean-sampling":
        n_sample_minority = min(target_stats.values())
        sampling_strategy = {key: n_sample_minority for key in target_stats.keys()}
    else:
        raise NotImplementedError

    return sampling_strategy


def _sampling_strategy_majority(y, sampling_type):
    """Returns sampling target by targeting the majority class only."""
    if sampling_type == "over-sampling":
        raise ValueError(
            "'sampling_strategy'='majority' cannot be used with over-sampler."
        )
    elif sampling_type == "under-sampling" or sampling_type == "clean-sampling":
        target_stats = _count_class_sample(y)
        class_majority = max(target_stats, key=target_stats.get)
        n_sample_minority = min(target_stats.values())
        sampling_strategy = {
            key: n_sample_minority
            for key in target_stats.keys()
            if key == class_majority
        }
    else:
        raise NotImplementedError

    return sampling_strategy


def _sampling_strategy_not_majority(y, sampling_type):
    """Returns sampling target by targeting all classes but not the
    majority."""
    target_stats = _count_class_sample(y)
    if sampling_type == "over-sampling":
        n_sample_majority = max(target_stats.values())
        class_majority = max(target_stats, key=target_stats.get)
        sampling_strategy = {
            key: n_sample_majority - value
            for (key, value) in target_stats.items()
            if key != class_majority
        }
    elif sampling_type == "under-sampling" or sampling_type == "clean-sampling":
        n_sample_minority = min(target_stats.values())
        class_majority = max(target_stats, key=target_stats.get)
        sampling_strategy = {
            key: n_sample_minority
            for key in target_stats.keys()
            if key != class_majority
        }
    else:
        raise NotImplementedError

    return sampling_strategy


def _sampling_strategy_not_minority(y, sampling_type):
    """Returns sampling target by targeting all classes but not the
    minority."""
    target_stats = _count_class_sample(y)
    if sampling_type == "over-sampling":
        n_sample_majority = max(target_stats.values())
        class_minority = min(target_stats, key=target_stats.get)
        sampling_strategy = {
            key: n_sample_majority - value
            for (key, value) in target_stats.items()
            if key != class_minority
        }
    elif sampling_type == "under-sampling" or sampling_type == "clean-sampling":
        n_sample_minority = min(target_stats.values())
        class_minority = min(target_stats, key=target_stats.get)
        sampling_strategy = {
            key: n_sample_minority
            for key in target_stats.keys()
            if key != class_minority
        }
    else:
        raise NotImplementedError

    return sampling_strategy


def _sampling_strategy_minority(y, sampling_type):
    """Returns sampling target by targeting the minority class only."""
    target_stats = _count_class_sample(y)
    if sampling_type == "over-sampling":
        n_sample_majority = max(target_stats.values())
        class_minority = min(target_stats, key=target_stats.get)
        sampling_strategy = {
            key: n_sample_majority - value
            for (key, value) in target_stats.items()
            if key == class_minority
        }
    elif sampling_type == "under-sampling" or sampling_type == "clean-sampling":
        raise ValueError(
            "'sampling_strategy'='minority' cannot be used with"
            " under-sampler and clean-sampler."
        )
    else:
        raise NotImplementedError

    return sampling_strategy


def _sampling_strategy_auto(y, sampling_type):
    """Returns sampling target auto for over-sampling and not-minority for
    under-sampling."""
    if sampling_type == "over-sampling":
        return _sampling_strategy_not_majority(y, sampling_type)
    elif sampling_type == "under-sampling" or sampling_type == "clean-sampling":
        return _sampling_strategy_not_minority(y, sampling_type)


def _sampling_strategy_dict(sampling_strategy, y, sampling_type):
    """Returns sampling target by converting the dictionary depending of the
    sampling."""
    target_stats = _count_class_sample(y)
    # check that all keys in sampling_strategy are also in y
    set_diff_sampling_strategy_target = set(sampling_strategy.keys()) - set(
        target_stats.keys()
    )
    if len(set_diff_sampling_strategy_target) > 0:
        raise ValueError(
            f"The {set_diff_sampling_strategy_target} target class is/are not "
            f"present in the data."
        )
    # check that there is no negative number
    if any(n_samples < 0 for n_samples in sampling_strategy.values()):
        raise ValueError(
            f"The number of samples in a class cannot be negative."
            f"'sampling_strategy' contains some negative value: {sampling_strategy}"
        )
    sampling_strategy_ = {}
    if sampling_type == "over-sampling":
        n_samples_majority = max(target_stats.values())
        class_majority = max(target_stats, key=target_stats.get)
        for class_sample, n_samples in sampling_strategy.items():
            if n_samples < target_stats[class_sample]:
                raise ValueError(
                    f"With over-sampling methods, the number"
                    f" of samples in a class should be greater"
                    f" or equal to the original number of samples."
                    f" Originally, there is {target_stats[class_sample]} "
                    f"samples and {n_samples} samples are asked."
                )
            if n_samples > n_samples_majority:
                warnings.warn(
                    f"After over-sampling, the number of samples ({n_samples})"
                    f" in class {class_sample} will be larger than the number of"
                    f" samples in the majority class (class #{class_majority} ->"
                    f" {n_samples_majority})"
                )
            sampling_strategy_[class_sample] = n_samples - target_stats[class_sample]
    elif sampling_type == "under-sampling":
        for class_sample, n_samples in sampling_strategy.items():
            if n_samples > target_stats[class_sample]:
                raise ValueError(
                    f"With under-sampling methods, the number of"
                    f" samples in a class should be less or equal"
                    f" to the original number of samples."
                    f" Originally, there is {target_stats[class_sample]} "
                    f"samples and {n_samples} samples are asked."
                )
            sampling_strategy_[class_sample] = n_samples
    elif sampling_type == "clean-sampling":
        raise ValueError(
            "'sampling_strategy' as a dict for cleaning methods is "
            "not supported. Please give a list of the classes to be "
            "targeted by the sampling."
        )
    elif sampling_type == "hybrid-sampling":
        for class_sample, n_samples in sampling_strategy.items():
            sampling_strategy_[class_sample] = n_samples
    else:
        raise NotImplementedError

    return sampling_strategy_


def _sampling_strategy_list(sampling_strategy, y, sampling_type):
    """With cleaning methods, sampling_strategy can be a list to target the
    class of interest."""
    if sampling_type != "clean-sampling":
        raise ValueError(
            "'sampling_strategy' cannot be a list for samplers "
            "which are not cleaning methods."
        )

    target_stats = _count_class_sample(y)
    # check that all keys in sampling_strategy are also in y
    set_diff_sampling_strategy_target = set(sampling_strategy) - set(
        target_stats.keys()
    )
    if len(set_diff_sampling_strategy_target) > 0:
        raise ValueError(
            f"The {set_diff_sampling_strategy_target} target class is/are not "
            f"present in the data."
        )

    return {
        class_sample: min(target_stats.values()) for class_sample in sampling_strategy
    }


def _sampling_strategy_float(sampling_strategy, y, sampling_type):
    """Take a proportion of the majority (over-sampling) or minority
    (under-sampling) class in binary classification."""
    type_y = type_of_target(y)
    if type_y != "binary":
        raise ValueError(
            '"sampling_strategy" can be a float only when the type '
            "of target is binary. For multi-class, use a dict."
        )
    target_stats = _count_class_sample(y)
    if sampling_type == "over-sampling":
        n_sample_majority = max(target_stats.values())
        class_majority = max(target_stats, key=target_stats.get)
        sampling_strategy_ = {
            key: int(n_sample_majority * sampling_strategy - value)
            for (key, value) in target_stats.items()
            if key != class_majority
        }
        if any([n_samples <= 0 for n_samples in sampling_strategy_.values()]):
            raise ValueError(
                "The specified ratio required to remove samples "
                "from the minority class while trying to "
                "generate new samples. Please increase the "
                "ratio."
            )
    elif sampling_type == "under-sampling":
        n_sample_minority = min(target_stats.values())
        class_minority = min(target_stats, key=target_stats.get)
        sampling_strategy_ = {
            key: int(n_sample_minority / sampling_strategy)
            for (key, value) in target_stats.items()
            if key != class_minority
        }
        if any(
            [
                n_samples > target_stats[target]
                for target, n_samples in sampling_strategy_.items()
            ]
        ):
            raise ValueError(
                "The specified ratio required to generate new "
                "sample in the majority class while trying to "
                "remove samples. Please increase the ratio."
            )
    else:
        raise ValueError(
            "'clean-sampling' methods do let the user specify the sampling ratio."
        )
    return sampling_strategy_


def check_sampling_strategy(sampling_strategy, y, sampling_type, **kwargs):
    """Sampling target validation for samplers.

    Checks that ``sampling_strategy`` is of consistent type and return a
    dictionary containing each targeted class with its corresponding
    number of sample. It is used in :class:`~imblearn.base.BaseSampler`.

    Parameters
    ----------
    sampling_strategy : float, str, dict, list or callable,
        Sampling information to sample the data set.

        - When ``float``:

            For **under-sampling methods**, it corresponds to the ratio
            :math:`\\alpha_{us}` defined by :math:`N_{rM} = \\alpha_{us}
            \\times N_{m}` where :math:`N_{rM}` and :math:`N_{m}` are the
            number of samples in the majority class after resampling and the
            number of samples in the minority class, respectively;

            For **over-sampling methods**, it correspond to the ratio
            :math:`\\alpha_{os}` defined by :math:`N_{rm} = \\alpha_{os}
            \\times N_{m}` where :math:`N_{rm}` and :math:`N_{M}` are the
            number of samples in the minority class after resampling and the
            number of samples in the majority class, respectively.

            .. warning::
               ``float`` is only available for **binary** classification. An
               error is raised for multi-class classification and with cleaning
               samplers.

        - When ``str``, specify the class targeted by the resampling. For
          **under- and over-sampling methods**, the number of samples in the
          different classes will be equalized. For **cleaning methods**, the
          number of samples will not be equal. Possible choices are:

            ``'minority'``: resample only the minority class;

            ``'majority'``: resample only the majority class;

            ``'not minority'``: resample all classes but the minority class;

            ``'not majority'``: resample all classes but the majority class;

            ``'all'``: resample all classes;

            ``'auto'``: for under-sampling methods, equivalent to ``'not
            minority'`` and for over-sampling methods, equivalent to ``'not
            majority'``.

        - When ``dict``, the keys correspond to the targeted classes. The
          values correspond to the desired number of samples for each targeted
          class.

          .. warning::
             ``dict`` is available for both **under- and over-sampling
             methods**. An error is raised with **cleaning methods**. Use a
             ``list`` instead.

        - When ``list``, the list contains the targeted classes. It used only
          for **cleaning methods**.

          .. warning::
             ``list`` is available for **cleaning methods**. An error is raised
             with **under- and over-sampling methods**.

        - When callable, function taking ``y`` and returns a ``dict``. The keys
          correspond to the targeted classes. The values correspond to the
          desired number of samples for each class.

    y : ndarray of shape (n_samples,)
        The target array.

    sampling_type : {{'over-sampling', 'under-sampling', 'clean-sampling'}}
        The type of sampling. Can be either ``'over-sampling'``,
        ``'under-sampling'``, or ``'clean-sampling'``.

    kwargs : dict
        Dictionary of additional keyword arguments to pass to
        ``sampling_strategy`` when this is a callable.

    Returns
    -------
    sampling_strategy_converted : dict
        The converted and validated sampling target. Returns a dictionary with
        the key being the class target and the value being the desired
        number of samples.

    """
    if sampling_type not in SAMPLING_KIND:
        raise ValueError(
            f"'sampling_type' should be one of {SAMPLING_KIND}. "
            f"Got '{sampling_type} instead."
        )

    if np.unique(y).size <= 1:
        raise ValueError(
            f"The target 'y' needs to have more than 1 class. "
            f"Got {np.unique(y).size} class instead"
        )

    if sampling_type in ("ensemble", "bypass"):
        return sampling_strategy

    if isinstance(sampling_strategy, str):
        if sampling_strategy not in SAMPLING_TARGET_KIND.keys():
            raise ValueError(
                f"When 'sampling_strategy' is a string, it needs"
                f" to be one of {SAMPLING_TARGET_KIND}. Got '{sampling_strategy}' "
                f"instead."
            )
        return OrderedDict(
            sorted(SAMPLING_TARGET_KIND[sampling_strategy](y, sampling_type).items())
        )
    elif isinstance(sampling_strategy, dict):
        return OrderedDict(
            sorted(_sampling_strategy_dict(sampling_strategy, y, sampling_type).items())
        )
    elif isinstance(sampling_strategy, list):
        return OrderedDict(
            sorted(_sampling_strategy_list(sampling_strategy, y, sampling_type).items())
        )
    elif isinstance(sampling_strategy, Real):
        if sampling_strategy <= 0 or sampling_strategy > 1:
            raise ValueError(
                f"When 'sampling_strategy' is a float, it should be "
                f"in the range (0, 1]. Got {sampling_strategy} instead."
            )
        return OrderedDict(
            sorted(
                _sampling_strategy_float(sampling_strategy, y, sampling_type).items()
            )
        )
    elif callable(sampling_strategy):
        sampling_strategy_ = sampling_strategy(y, **kwargs)
        return OrderedDict(
            sorted(
                _sampling_strategy_dict(sampling_strategy_, y, sampling_type).items()
            )
        )


SAMPLING_TARGET_KIND = {
    "minority": _sampling_strategy_minority,
    "majority": _sampling_strategy_majority,
    "not minority": _sampling_strategy_not_minority,
    "not majority": _sampling_strategy_not_majority,
    "all": _sampling_strategy_all,
    "auto": _sampling_strategy_auto,
}


def _deprecate_positional_args(f):
    """Decorator for methods that issues warnings for positional arguments

    Using the keyword-only argument syntax in pep 3102, arguments after the
    * will issue a warning when passed as a positional argument.

    Parameters
    ----------
    f : function
        function to check arguments on.
    """
    sig = signature(f)
    kwonly_args = []
    all_args = []

    for name, param in sig.parameters.items():
        if param.kind == Parameter.POSITIONAL_OR_KEYWORD:
            all_args.append(name)
        elif param.kind == Parameter.KEYWORD_ONLY:
            kwonly_args.append(name)

    @wraps(f)
    def inner_f(*args, **kwargs):
        extra_args = len(args) - len(all_args)
        if extra_args > 0:
            # ignore first 'self' argument for instance methods
            args_msg = [
                f"{name}={arg}"
                for name, arg in zip(kwonly_args[:extra_args], args[-extra_args:])
            ]
            warnings.warn(
                f"Pass {', '.join(args_msg)} as keyword args. From version 0.9 "
                f"passing these as positional arguments will "
                f"result in an error",
                FutureWarning,
            )
        kwargs.update({k: arg for k, arg in zip(sig.parameters, args)})
        return f(**kwargs)

    return inner_f



"""
--------------------------------------------------------------------------
The `self_paced_ensemble.self_paced_ensemble` module implement a 
self-paced Ensemble (SPE) Classifier for binary class-imbalanced learning.
    
Self-paced Ensemble (SPE) is an ensemble learning framework for massive highly 
imbalanced classification. It is an easy-to-use solution to class-imbalanced 
problems, features outstanding computing efficiency, good performance, and wide 
compatibility with different learning models.

See https://github.com/ZhiningLiu1998/self-paced-ensemble.

Reference:
Liu Z, Cao W, Gao Z, et al. Self-paced ensemble for highly imbalanced 
massive data classification[C]//2020 IEEE 36th International Conference 
on Data Engineering (ICDE). IEEE, 2020: 841-852.
--------------------------------------------------------------------------
"""

from ._self_paced_ensemble import SelfPacedEnsembleClassifier

__all__ = [
    "SelfPacedEnsembleClassifier",
]
